# 04-Python变量

[TOC]

## 一、变量概念

​	Python中会用到许多数据，为了方便操作，需要把这些数据分别用一个简单的名字代表，方便在接下来的程序中引用。变量就是代表某个数据(值)的名称。简单点说变量就是给数据起个名

## 二、变量名称

**命名规定**

​	由字母数字下划线组成的，且不能以数字开头，不能使用关键字，区分大小写。

**关键字：**
不要以关键字命名变量

```python
>>> import keyword
>>> keyword.kwlist
    ['False', 'None', 'True', 'and', 'as', 'assert', 'break', 'class', 'continue', 'def', 'del', 'elif', 'else', 'except', 'finally', 'for', 'from', 'global', 'if', 'import', 'in', 'is', 'lambda', 'nonlocal', 'not', 'or', 'pass', 'raise', 'return', 'try', 'while', 'with', 'yield']
```

可以用以下方法验证，返回 True 就是 Python 的关键字

```python
>>> keyword.iskeyword('del')
    True
```

**命名规范：**

```
模块：小写字母，单词之间用_分割，比如：ad_stats.py

包名：同模块命名规范

类名：单词首字母大写，比如：AdStats ConfigUtil

全局变量：大写字母，单词之间用_分割，比如：NUMBER  COLOR_WRITE

普通变量：小写字母，单词之间用_分割，比如：this_is_a_var


实例变量：以_开头，其他和普通变量一样，比如：price  _instance_var

私有实例变量（外部访问会报错）：以__开头（2个下划线），其他和普通变量一样，比如：__private_var

专有变量： __开头，__结尾，一般为python的自有变量，不要以这种方式命名，比如：__doc__   __class
```

## 三、变量赋值

变量赋值是变量声明和定义的过程 
**单个变量赋值：**

```python
#!/usr/bin/python3.8
counter = 100  # 赋值整型
miles = 1000.0 # 赋值浮点型
name = "John"  # 赋值字符串

print(counter)
print(miles)
print(name)       
```

**多个变量赋值**
    Python允许同时为多个变量赋值。例如： 

```python
a = b = c = 1
# 创建一个整型对象，值为1，三个变量被分配到相同的内存空间上。
```

​    同时为多个变量赋不同的值。例如：

```python
a, b, c = 1, 2, "john"
# 两个整型对象1和2分配给变量a和b，字符串对象"john"分配给变量c。
```

## 四、存储方式

 1、一般编程语言变量存储数据的方式
    	变量是计算机内存中的一块区域，变量可以存储规定范围内的值，而且值是可变的。
    	在创建变量时会在内存中开辟一个空间。基于变量的数据类型，解释器会分配指定内存，并决定什么数据可被存储在内存中。因此，变量可以指定不同的数据类型，这些变量可以存储整数，小数或字符。 
    	如c语言在声明一个变量a之后，会在内存中开辟出一块对应的空间，在此空间中可以存储不同的值，也就是给变量赋予不同的值

2、Python变量在内存中存储值得方式和其他编程语言不同
    	在Python中，变量名没有类型，但对象有；变量名只是对对象的引用（内部实现为指针）       

​    	Python中是以数据为主，变量a只是相当于一个内存空间的标签，a=1开辟一块空间存储1,之后重新复制a=2是重新开辟出新的空间存储2，变量名称a换了个位置指向新空间中的2

​    	同样的地址空间可以有两个或多个标签，比如a=1,b=1实际上是a和b指向同一个地址空间

**地址空间**

​	查看变量指向内存地址空间的地址：使用id()函数

```python
>>> a=1
>>> id(a)
	19882304
>>> b=1
>>> id(b)
	19882304 
	# 发现同一个值赋值给不同变量，实际地址空间未发生变化，只是标签发生了变化
```

**对象类型**

对象都有不同的类型，用 type()函数查看

```
In [9]: type(1)
Out[9]: int
1  # 值，对象本身
```

**对象比较**

```python
# 通常用 `is` 或者`==`来判断两个对象的标识是否一样，即判断两个变量绑定的是不是同一个对象
a = 'wing'
b = a
a is b
```

```python
x = 'wing'
y = 'wing'
x is y
x == y
```

## 五、引用计数

在 Python 中，不用关心内存溢出问题，Python 已经帮忙实现了内存管理。

**注意：sys.getrefcount()获取的引用计数在2.7没问题，3.0内只对字符串生效，获取数值类型的计数不准确**

**引用计数器**
    Python内部记录着所有使用中的对象有多少引用。一个内部跟踪变量，称为一个引用计数器。当对象被创建时，就创建了一个引用计数，当这个对象不再需要时，也就是这个对象的引用计数变为0时，它被垃圾回收。(这个只是形象的说一下，并不是严格的100%正确，但是通俗的理解往往是最好的学习方式)

**增加引用计数**
    1、当对象被创建并赋值给变量时，该对象的引用技术就被设置为1。
    2、当同一个对象的应用或是对象又被赋值给其他变量时，或者作为参数传递给函数，方法或类实例时，或者被赋值为一个窗口对象的成员时，该对象的一个新的引用，或者称作别名，就被创建，则该对象的引用计数自动加1

**减少引用计数**
    1、当对象的引用被销毁时，引用计数会减少。最明显的例子就是当引用离开其作用范围时，这种情况最经常出现在函数运行结束时，所有局部变量都被自动销毁，对象的引用计数也就随之减少。

​    2、当变量被赋值给另外一个对象时，源对象的引用技术也会自动减1

​	3、其他造成对象的引用计数减少的方式包括使用del语句删除一个变量

**Python内部的引用计数使用sys.getrefcount获取** 

例子：references

```python
>>> import sys
>>> a="ab"
>>> sys.getrefcount("ab")
    3             # 第一次结果为3
>>> b="ab"
>>> sys.getrefcount("ab")
    4             # 第二次结果+1
>>> b=0   		  # b引用了其他的对象（0），对于"ab"来讲就取消了一个引用
>>> sys.getrefcount("ab")
    3             # 结果在上次引用的基础上-1
    
   
```

注意：在交互式解释器中`带空格`的对象引用次数永远为3，但是在脚本中回归正常，例如：

```python
#!/usr/bin/env python
import sys
print sys.getrefcount("ab cd")
a="ab cd"
print sys.getrefcount("ab cd")
b="ab cd"
print sys.getrefcount("ab cd")
c=b
print sys.getrefcount("ab cd")
```

**垃圾收集**
    不再被使用的内存会被一种称为垃圾收集的机制释放。虽然解释器跟踪对象的引用计数，但是垃圾收集器负责释放内存。垃圾收集器是一块独立的代码，它用来寻找引用计数为0的对象，也负责检查那些虽然引用计数大于0但也该被销毁的对象。特定情形会导致循环引用。
    一个循环引用发生在当你有至少两个对象互相引用时，也就是所说的引用都消失时，这些引用仍然存在，这说明只靠引用计数是不够的。Python的垃圾收集器实际上是**一个引用计数器和一个循环垃圾收集器**。当一个对象的引用计数变为0，解释器会暂停，释放掉这个对象和仅有这个对象可访问的其他对象，作为引用计数的补充，垃圾收集器也会留心被分配的总量很大(以及未通过引用计数销毁的那些)  的对象。在这种情况下，解释器会暂停下来，试图清理所有为引用的循环。

## 六、变量互传

Python和Shell变量互相传递的方法

shell==>python:

```python
import os
var=os.popen('df -h').read( ) 
print(var)

执行结果：
[root@workstation day01]# /usr/bin/python3.8 "/root/Development/python_code/day01/hello.py"
Filesystem           Size  Used Avail Use% Mounted on
devtmpfs             3.9G     0  3.9G   0% /dev
tmpfs                3.9G   57M  3.9G   2% /dev/shm
tmpfs                3.9G   10M  3.9G   1% /run
tmpfs                3.9G     0  3.9G   0% /sys/fs/cgroup
/dev/mapper/cl-root   44G  7.4G   37G  17% /
/dev/nvme0n1p1       976M  234M  675M  26% /boot
tmpfs                795M  1.2M  794M   1% /run/user/42
tmpfs                795M  6.9M  788M   1% /run/user/0
/dev/sr0             7.7G  7.7G     0 100% /run/media/root/CentOS-8-2-2004-x86_64-dvd
```

python==>shell:

```python
import os 
var=123
os.environ['var']=str(var)  #environ的键值必须是字符串

os.system('echo $var')         
```

```python
不用转变量的方式替代上面的写法
In [9]: c="/tmp/hello"
In [10]: os.system("mkdir"+" "+c)
```



# **用shell编写自己的wc命令**













