# Python面向对象

## 面向对象

面向对象编程（Object-Oriented Programming，OOP）是一种程序设计范式，它以对象为中心，将数据和操作数据的方法封装在一起。在Python中，一切皆为对象，包括整数、字符串、函数等。﻿

在Python中，类和OOP 都不是日常编程所必需的。尽管它从一开始设计就是面向对象的,并且结构上支持 OOP,但 Python 没有限定或要求你在你的应用中写 OOP 的代码。
[![file://C:\Users\86186\AppData\Local\Temp\.2IK5Q0\1.png](assets/1.png)]()[![file://C:\Users\86186\AppData\Local\Temp\.2IK5Q0\2.png](assets/2.png)]()

## 相关概念

**类(Class)**

​	用来描述具有相同的属性和方法的对象的集合。它定义了该集合中每个对象所共有的属性和方法。对象是类的实例。

**方法**

​	类中定义的函数。

**类变量**

​	类变量在整个实例化的对象中是公用的。类变量定义在类中且在函数体之外。类变量通常不作为实例变量使用。

**实例变量**

​	定义在方法中的变量，只作用于当前实例

**数据成员**

​	类变量或者实例变量，用于处理类及其实例对象的相关的数据。

**方法重写**

​	如果从父类继承的方法不能满足子类的需求，可以对其进行改写，这个过程叫方法的覆盖（override），也称为方法的重写。

**继承**

​	即一个派生类（derived class）继承基类（base class）的字段和方法。继承也允许把一个派生类的对象作为一个基类对象对待。例如，有这样一个设计 ：一个Dog类型的对象派生自Animal类，这是模拟"是一个（is-a）"关系（例如，Dog是一个Animal）。

**实例化**

​	创建一个类的实例，类的具体对象。

**对象**

​	通过类定义的数据结构实例。对象包括两个数据成员（类变量和实例变量）和方法。

## 创建类和实例

使用class语句来创建一个新类，class之后为类的名称并以冒号结尾

```python
class ClassName:
   '类的帮助信息'      # 类文档字符串，类的帮助信息可以通过ClassName.__doc__查看
   class_suite        # 类体，class_suite 由类成员，方法，数据属性组成。
```

例

```python
#!/usr/bin/env python
class Employee:   					# 创建类
    '所有员工的基类'
    empCount = 0 
    
    # 是一个类变量，它的值将在这个类的所有实例之间共享。可在内部类或外部类使用Employee.empCount访问。也叫 静态变量 静态数据 静态字段 静态属性 类数据属性，静态变量在内存中只保存一份，实例变量在每个对象中都保存一份
    def __init__(self, name, salary): 
    # 是一种特殊的方法，被称为类的构造函数或初始化方法，当创建了这个类的实例时就会调用该方法
        self.name = name    # 实例属性，也叫普通变量或实例变量
        self.salary = salary
        Employee.empCount += 1

    def displayCount(self):
        print("Total Employee {}".format(Employee.empCount))

    def displayEmployee(self):
        print("Name : ", self.name, ", Salary: ", self.salary)
'''   
创建实例对象
	要创建一个类的实例，可以使用类的名称，并通过`__init__`方法接收参数。
'''
emp1=Employee("wing",2000)  		# 创建 Employee 类的第一个对象
emp2=Employee("lilei",3000)			# 创建 Employee 类的第二个对象

# 访问属性  使用.来访问对象的属性
emp1.displayEmployee()
emp2.displayEmployee()
print("Total Employee {}".format(Employee.empCount))

emp1.age = 7   # 添加一个 'age' 属性
emp1.age = 8   # 修改     'age' 属性
del emp1.age   # 删除     'age' 属性

输出结果
[root@www test]# python test.py 
Name :  wing , Salary:  2000
Name :  lilei , Salary:  3000
Total Employee 2
```

函数方式访问属性

```python
访问对象的属性
getattr(obj, name[, default])
例子
    getattr(emp1, 'age')      # 返回 'age' 属性的值

检查是否存在一个属性    
hasattr(obj,name)
例子
    hasattr(emp1, 'age')      # 如果存在 'age' 属性返回 True。

设置一个属性。如果属性不存在，会创建一个新属性
setattr(obj,name,value)
例子
	setattr(emp1, 'age', 8)  # 添加属性 'age' 值为 8

删除属性
delattr(obj, name)
例
	delattr(emp1, 'age')     # 删除属性 'age'
```

## 类属性与方法

**类的私有属性**(私有字段)

`__private_attrs`

​	两个下划线开头，声明该属性为私有，不能在类地外部被使用或直接访问。

​	在类内部的方法中使用时 self.__private_attrs

**类的方法**
	在类地内部，使用def关键字可以为类定义一个方法，与一般函数定义不同，类方法必须包含参数self,且为第一个参数

**类的私有方法**

`__private_method`

​	两个下划线开头，声明该方法为私有方法，不能在类地外部调用。

​	在类的内部调用 self.__private_methods 

实例

```python
#!/usr/bin/python
class JustCounter:
	__secretCount = 0   # 私有变量
	publicCount = 0     # 公开变量

	def count(self):
		self.__secretCount += 1
		self.publicCount += 1
		print self.__secretCount

counter = JustCounter()
counter.count()
counter.count()
print counter.publicCount
print counter.__secretCount   # 报错，实例不能访问私有变量Python 通过改变名称来包含类名:
1
2
2
Traceback (most recent call last):
  File "test.py", line 17, in <module>
    print counter.__secretCount     # 报错，实例不能访问私有变量
AttributeError: JustCounter instance has no attribute '__secretCount'

Python不允许实例化的类访问私有数据，但你可以使用 object._className__attrName 访问属性，将如下代码替换以上代码的最后一行代码：
.........................
print counter._JustCounter__secretCount
执行结果：
1
2
2
2  
```

**特殊类属性(Python内置类属性)：**
   C.__dict__ :类C的属性（包含一个字典，由类的数据属性组成） 
   C.__doc__ :类C的文档字符串 
   C.__name__:类C类名(字符串) 
   C.__module__:类C定义所在的模块（类的全名是'__main__.className'，如果类位于一个导入
模块mymod中，那么className.__module__ 等于 mymod） 
   C.__bases__ : 类C的所有父类构成元素（包含了一个由所有父类组成的元组） 
   C.__class__:实例对应的类(仅新式类中)

根据上面定义的类 MyClass,有如下结果:
     \>>> MyClass.__name__
    'MyClass'

​    \>>> MyClass.__doc__
​    'MyClass class definition'

​    \>>> MyClass.__bases__
​    (<type 'object'>,)

​    \>>> print MyClass.__dict__
​    {'__doc__': None, 'myVersion': 1, 'showMyVersion':
​    <function showMyVersion at 950ed0>, '__module__': '__main__'}

​    \>>> MyClass.__module__
​    '__main__'
​    \>>> MyClass.__class__
​    <type 'type'>

**__name__**
    是给定类的字符名字。
    它适用于那种只需要字符串(类对象的名字),而非类对象本身的情况。一些内建的类型也有这个属性,我们将会用
到其中之一来展示__name__字符串的益处。

​    类型对象是一个内建类型的例子,它有__name__的属性。回忆一下,type()返回被调用对象的类型。
​    我们可以使用类型对象的__name__属性来取得相应的字符串名。

如下例示:
    \>>> stype = type('What is your quest?')
    \>>> stype                       # stype is a type object stype 是一个类型对象
    <type 'string'>
    \>>> stype.__name__      # get type as a string 得到类型名(字符串表示)
    'string'

​    \>>> type(3.14159265)    # also a type object 又一个类型对象
​    <type 'float'>
​    \>>> type(3.14159265).__name__   # get type as a string 得到类型名(字符串表示) 
​    'float'


**__doc__**
    是类的文档字符串,与函数及模块的文档字符串相似,必须紧随头行(header line) 后的字符串。
    文档字符串不能被派生类继承,也就是说派生类必须含有它们自己的文档字符串。


**__dict__**
    包含一个字典,由类的数据属性组成。
    访问一个类属性的时候,Python 解释器将会搜索字典以得到需要的属性。
    如果在__dict__中没有找到,将会在基类的字典中进行搜索, 采用“"度优先搜索”"序。
    基类集的搜索是按顺序的,从左到右,按其在类定义时,定义父类参数时的顺序。
    对类的修改会仅影响到此类的字典;基类的__dict__属性不会被改动的


**__module__**
    Python 支持模块间的类继承。为更清晰地对类进行描述,1.5 版本中引入了__module__, 这样类名就完全由模
块名所限定。看一下下面的例子:
    \>>> class C(object): 
    ...         pass
    ...
    \>>> C
    <class __main__.C at 0x53f90>

​    \>>> C.__module__
​    '__main__'

​    类 C 的全名是"__main__.C",比如,source_module.class_name。
​    如果类 C 位于一个导入的模块中,如 mymod,像下面的: 
​    \>>> from mymod import C 

​    \>>> C
​    <class mymod.C at 0x53ea0> 

​    \>>> C.__module__
​    'mymod'

​    在以前的版本中,没有特殊属性__module__,很难简单定位类的位置,因为类没有使用它们的全名。

**__class__**
    最后,由于类型和类的统一性,当访问任何类的__class__属性时,你将发现它就是一个类型对象的实例。换句话说,一个类已是一种类型了。因为经典类并不认同这种等价性(一个经典类是一 个类对象,一个类型是一个类型对象),对这些对象来说,这个属性并未定义。
**__class__到底是什么意思？看例子**
    \# cat a.py
    \#!/usr/bin/python
    \# -*- coding: UTF-8 -*-
    class A:
        def __init__(self,url):
            self.url = url
        def out(self):
            return self.url

​    a = A('news.163.com')
​    print a.out()

​    b = a.__class__('www.bccn.net')
​    print b.out()

print A
    print a.__class__

​    \# python a.py
​    news.163.com
​    www.bccn.net
​    __main__.A
​    __main__.A

**可以看出a.__class__就等效于a的类A**

内置类属性调用实例：
    \#!/usr/bin/python
    class Employee:
       '所有员工的基类'
       empCount = 0

​       def __init__(self, name, salary):
​          self.name = name
​          self.salary = salary
​          Employee.empCount += 1

​       def displayCount(self):
​         print "Total Employee %d" % Employee.empCount

​       def displayEmployee(self):
​          print "Name : ", self.name,  ", Salary: ", self.salary

​    print "Employee.__doc__:", Employee.__doc__
​    print "Employee.__name__:", Employee.__name__
​    print "Employee.__module__:", Employee.__module__
​    print "Employee.__bases__:", Employee.__bases__
​    print "Employee.__dict__:", Employee.__dict__

​    输出结果：
​    Employee.__doc__: 所有员工的基类
​    Employee.__name__: Employee
​    Employee.__module__: __main__
​    Employee.__bases__: ()
​    Employee.__dict__: {'__module__': '__main__', 'displayCount': <function displayCount at 0x10a939c80>, 'empCount': 0, 'displayEmployee': <function displayEmployee at 0x10a93caa0>, '__doc__': '\xe6\x89\x80\xe6\x9c\x89\xe5\x91\x98\xe5\xb7\xa5\xe7\x9a\x84\xe5\x9f\xba\xe7\xb1\xbb', '__init__': <function __init__ at 0x10a939578>}

## 类的继承

面向对象的编程带来的主要好处之一是代码的重用，实现这种重用的方法之一是通过继承机制。
继承完全可以理解成类之间的类型和子类型关系。

继承语法
    class 派生类名（基类名）：

python继承中的特点：
    1：在继承中基类的构造（__init__()方法）不会被自动调用，它需要在其派生类的构造中亲自专门调用。
    2：在调用基类的方法时，需要加上基类的类名前缀，且需要带上self参数变量。区别于在类中调用普通函数时并不需要带上self参数
    3：Python总是首先查找对应类型的方法，如果它不能在派生类中找到对应的方法，它才开始到基类中逐个查找。（先在本类中查找调用的方法，找不到才去基类中找）。

多重继承
如果在继承元组中列了一个以上的类，那么它就被称作"多重继承" 。
语法：
派生类的声明，与他们的父类类似，继承的基类列表跟在类名之后，如下所示：
class SubClassName (ParentClass1[, ParentClass2, ...]):
   'Optional class documentation string'
   class_suite

实例：
    \#!/usr/bin/python
    class Parent:        # 定义父类
       parentAttr = 100
       def __init__(self):
          print "调用父类构造函数"

​       def parentMethod(self):
​          print '调用父类方法'

​       def setAttr(self, attr):
​          Parent.parentAttr = attr

​       def getAttr(self):
​          print "父类属性 :", Parent.parentAttr

​    class Child(Parent): # 定义子类
​       def __init__(self):
​          print "调用子类构造方法"

​       def childMethod(self):
​          print '调用子类方法 child method'

​    c = Child()              # 实例化子类
​    c.childMethod()      # 调用子类的方法
​    c.parentMethod()   # 调用父类方法
​    c.setAttr(200)        # 再次调用父类的方法
​    c.getAttr()              # 再次调用父类的方法

以上代码执行结果如下：
    调用子类构造方法
    调用子类方法 child method
    调用父类方法
    父类属性 : 200

继承多个类
class A:            # 定义类 A
.....

class B:            # 定义类 B
.....

class C(A, B):   # 继承类 A 和 B
.....

使用issubclass()或者isinstance()方法来检测。
issubclass()布尔函数判断一个类是另一个类的子类或者子孙类，语法： 
    issubclass(sub,sup)

例子:    
    def foo():  
        pass                
    print issubclass(foo.__class__, object) 

​    结果:        
​    True

​    上述代码说明了Python 中的函数是 object 的子类
​    
isinstance(obj, Class) 布尔函数如果obj是Class类的实例对象或者是一个Class子类的实例对象则返回true。

**方法重写**
如果你的父类方法的功能不能满足你的需求，你可以在子类重写你父类的方法：

实例：
\#!/usr/bin/python
class Parent:        # 定义父类
   def myMethod(self):
      print '调用父类方法'

class Child(Parent): # 定义子类
   def myMethod(self):
      print '调用子类方法'

c = Child()               # 子类实例
c.myMethod()         # 子类调用重写方法

输出结果：
    调用子类方法

基础重载方法
下表列出了一些通用的功能，你可以在自己的类重写：

| 序号 | 方法 描述 简单的调用                                         |
| ---- | ------------------------------------------------------------ |
| 1    | __init__ ( self [,args...] ) 构造函数 简单的调用方法: obj = className(args) |
| 2    | __del__( self ) 析构方法, 删除一个对象 简单的调用方法 : del obj |
| 3    | __repr__( self ) 转化为供解释器读取的形式 简单的调用方法 : repr(obj) |
| 4    | __str__( self ) 用于将值转化为适于人阅读的形式 简单的调用方法 : str(obj) |
| 5    | __cmp__ ( self, x ) 对象比较 简单的调用方法 : cmp(obj, x)    |



运算符重载
Python同样支持运算符重载，实例如下：
\#!/usr/bin/python
class Vector:
   def __init__(self, a, b):
      self.a = a
      self.b = b

   def __str__(self):
      return 'Vector (%d, %d)' % (self.a, self.b)

   def __add__(self,other):
      return Vector(self.a + other.a, self.b + other.b)

v1 = Vector(2,10)
v2 = Vector(5,-2)
print v1 + v2
以上代码执行结果如下所示:
Vector(7,8)

**类、实例和其他对象的内建函数**
**issubclass()**
issubclass() 布尔函数判断一个类是另一个类的子类或子孙类。
语法:
    issubclass(sub, sup)

issubclass() 返回True的情况:
   \1.  给出的子类sub确实是父类sup的一个子类(反之,则为False)。 
   \2.  这个函数也允许"不严格"的子类,意味着,一个类可视为其自身的子类,所以,这个函数如果当 sub 就是 sup,或者从 sup 派生而来,则返回 True。(一个"严格的"子类是严格意义上的从一个类派生而来的子类。)
   \3.  从 Python 2.3 开始,issubclass()的第二个参数可以是可能的父类组成的 tuple(元组),这时, 只要第一个参数是给定元组中任何一个候选类的子类时,就会返回 True。

**isinstance()**
isinstance() 布尔函数在判定一个对象是否是另一个给定类的实例时,非常有用。

语法:
    isinstance(obj1, obj2)
    isinstance()在 obj1 是类 obj2 的一个实例,或者是 obj2 的子类的一个实例时,返回 True (反之,则为 False)

例子:
    \>>> class C1(object): pass
    \>>> class C2(object): pass
    \>>> c1 = C1()
    \>>> c2 = C2()
    \>>> isinstance(c1, C1) True
    \>>> isinstance(c2, C1) False
    \>>> isinstance(c1, C2) False
    \>>> isinstance(c2, C2) True
    \>>> isinstance(C2, c2) 
    Traceback (innermost last):
    File "<stdin>", line 1, in ?
    isinstance(C2, c2)
    TypeError: second argument must be a class

注意:第二个参数应当是类;不然,你会得到一个 TypeError。但如果第二个参数是一个类型对象,则不会出现异常。这是允许的,因为你也可以使用 isinstance()来检查一个对象 obj1 是否是 obj2 的类型,比如:
    \>>> isinstance(4, int)
    True
    \>>> isinstance(4, str)
    False
    \>>> isinstance('4', str)
    True

isinstance()也可以使用一个元组(tuple)作为第二个参数。如果第一个参数是第二个参数中给定元组的任何一个候选类型或类的实例时,就会返回 True。

**hasattr(), getattr(),setattr(), delattr()**
*attr()系列函数可以在各种对象下工作,不限于类(class)和实例(instances),因为在类和实例中使用极其频繁,就在这里列出来了。

hasattr()函数是 Boolean 型的,它的目的就是为了决定一个对象是否有一个特定的属性,一般用于访问某属性前先作一下检查。

getattr()取得对象的属性,会在你试图读取一个不存在的属性时,引发 AttributeError 异常,除非给出那个可选的默认参数。

setattr()将要么加入一个新的属性,要么取代一个已存在的属性。

delattr()函数会从一个对象中删除属性。

下面一些例子使用到了*attr()系列函数:
\>>> class myClass(object): 
...         def __init__(self): 
...             self.foo = 100 
\>>> myInst = myClass()
\>>> hasattr(myInst, 'foo')
True
\>>> getattr(myInst, 'foo')
100
\>>> hasattr(myInst, 'bar') 
False
\>>> getattr(myInst, 'bar') 
Traceback (most recent call last):
File "<stdin>", line 1, in ?
getattr(myInst, 'bar')
AttributeError: myClass instance has no attribute 'bar'
\>>> getattr(c, 'bar', 'oops!')
'oops!'
\>>> setattr(myInst, 'bar', 'my attr')
\>>> dir(myInst)
['__doc__', '__module__', 'bar', 'foo']
\>>> getattr(myInst, 'bar')       # same as myInst.bar 
'my attr'
\>>> delattr(myInst, 'foo')
\>>> dir(myInst)
['__doc__', '__module__', 'bar']
\>>> hasattr(myInst, 'foo')
False

**dir()**
用 dir()列出一个模块所有属性的信息, dir()还可以用在对象上。
dir()提供的信息比以前更加详尽。根据文档,"除了实例变量名和常用方法外,它还显示那些通过特殊标记来调用的方法,像__iadd__(+=),__len__(len()), __ne__(!=)。
dir()作用在实例上(经典类或新式类)时,显示实例变量,还有在实例所在的类及所有它的基类中定义的方法和类属性.
dir()作用在类上(经典类或新式类)时,则显示类以及它的所有基类的__dict__中的内容。 但它不会显示定义在元类(metaclass)中的类属性.
dir()作用在模块上时,则显示模块的__dict__的内容.
dir()不带参数时,则显示调用者的局部变量.

**super()**
super()函数的目的就是帮助程序员找出相应的父类, 然后方便调用相关的属性。一般情况下,程序员可能仅仅采用非绑定方式调用祖先类方法。使用 super()可以简化搜索一个合适祖先的任务,并且在调用它时,替你传入实例或类型对象。
每个定义的类,都有一个名为__mro__的属性,它是一个元组,按照他们被搜索时的顺序,列出了备搜索的类。
语法如下:
    super(type[, obj])
给出 type,super()"返回此 type 的父类"。如果你希望父类被绑定,你可以传入 obj 参数(obj 必须是 
type 类型的).否则父类不会被绑定。obj 参数也可以是一个类型,但它应当是 type 的一个子类。通常,当给出 obj 时:
   \1.  如果obj是一个实例,isinstance(obj,type)就必须返回True
   \2.  如果obj是一个类或类型,issubclass(obj,type)就必须返回True

事实上,super()是一个工厂函数,它创造了一个 super object,为一个给定的类使用__mro__ 去查找相应的父类。很明显,它从当前所找到的类开始搜索 MRO。
super()的主要用途,是来查找父类的属性,比如, super(MyClass,self).__init__()。如果你没有执行这样的查找,你可能不需要使用 super()。

**vars()**
vars()内建函数与 dir()相似,只是给定的对象参数都必须有一个__dict__属性。vars()返回一个字典,它包含了对象存储于其__dict__中的属性(键)及值。如果提供的对象没有这样一个属性, 则会引发一个 TypeError 异常。如果没有提供对象作为 vars()的一个参数,它将显示一个包含本地名字空间的属性(键)及其值的字典,也就是locals()。

例子,使用类实例调用 vars():
    \>>> class C(object):
    ...         pass
    \>>> c = C()
    \>>> c.foo = 100
    \>>> c.bar = 'Python'
    \>>> c.__dict__
    {'foo': 100, 'bar': 'Python'}
    \>>> vars(c)
     {'foo': 100, 'bar': 'Python'}

**Python新式类和经典类**
Python从2.2开始，引入了 new style class（新式类）
Python 2.x中默认都是经典类，只有显式继承了object才是新式类
Python 3.x中默认都是新式类，不必显式的继承object

新式类跟经典类的差别主要是以下几点:
• 新式类对象可以直接通过__class__属性获取自身类型:type 
• 继承搜索的顺序发生了改变
• 新式类增加了__slots__内置属性, 可以把实例属性的种类锁定到__slots__规定的范围之中
• 新式类增加了__getattribute__方法

\1. 新式类对象可以直接通过__class__属性获取自身类型:type 
\#cat test1.py
\#/usr/bin/env python2.7
    class E:    
    \#经典类  
        pass  

​    class E1(object):    
​    \#新式类  
​        pass  

​    e = E()  
​    print "经典类"  
​    print e  
​    print type(e)  
​    print e.__class__  

​    print "新式类"  
​    e1 = E1()  
​    print e1  
​    print e1.__class__  
​    print type(e1)

输出结果：
    经典类  
    <__main__.E instance at 0x0000000002250B08>  
    <type 'instance'>  
    __main__.E  

​    新式类  
​    <__main__.E1 object at 0x0000000002248710>  
​    <class '__main__.E1'>  
​    <class '__main__.E1'> 

​    E1是定义的新式类。那么输出e1的时候，不论是type(e1)，还是e1.__class__都是输出的<class '__main__.E1'>。

\2. 继承搜索的顺序发生了改变
经典类多继承属性搜索顺序: 先深入继承树左侧，再返回，开始找右侧
新式类多继承属性搜索顺序: 先水平搜索，然后再向上移动
\#cat test2.py
    class A(object):    
        '新式类 ,作为所有类的基类'
        def foo(self):    
            print "class A"   

​    class A1():    
​        '经典类,作为所有类的基类 '
​        def foo(self):    
​            print "class A1"    

​    class C(A):    
​        pass  

​    class C1(A1):    
​        pass  

​    class D(A):    
​        def foo(self):    
​            print "class D"    

​    class D1(A1):    
​        def foo(self):    
​            print "class D1"    

​    class E(C, D):    
​        pass  

​    class E1(C1, D1):    
​        pass  

​    e = E()  
​    e.foo()  

​    e1 = E1()  
​    e1.foo() 

输出结果：
    class D  
    class A1  

因为A新式类,对于继承A类都是新式类，首先要查找类E中是否有foo()，如果没有则按顺序查找C->D->A。它是一种广度优先查找方式。

因为A1经典类,对于继承A1类都是经典类，首先要查找类E1中是否有foo()，如果没有则按顺序查找C1->A1->D1。它是一种深度优先查找方式。

\3. 新式类增加了__slots__内置属性, 可以把实例属性的种类锁定到__slots__规定的范围之中。

比如只允许对A实例添加name和age属性:
\#cat test3.py
    class A(object):    
        __slots__ = ('name', 'age')   

​    class A1():    
​        __slots__ = ('name', 'age')   

​    a1 = A1()  
​    a = A()  

​    a1.name1 = "a1"  
​    a.name1 = "a" 

A是新式类添加了__slots__ 属性,所以只允许添加 name age
A1经典类__slots__ 属性没用

执行结果：
     Traceback (most recent call last):  
       File "t.py", line 13, in <module>  
         a.name1 = "a"  
     AttributeError: 'A' object has no attribute 'name1'  
所以a.name是会出错的

通常每一个实例都会有一个__dict__属性，用来记录实例中所有的属性和方法，也是通过这个字典，可以让实例绑定任意的属性。
__slots__属性作用就是，当类C有比较少的变量，而且拥有__slots__属性时，类C的实例就没有__dict__属性，而是把变量的值存在一个固定的地方。如果试图访问一个__slots__中没有的属性，实例就会报错。
这样操作的好处：__slots__属性虽然令实例失去了绑定任意属性的便利，但是因为每一个实例没有__dict__属性，却能有效节省每一个实例的内存消耗，有利于生成小而精干的实例。

\4. 新式类增加了__getattribute__方法
\#cat test4.py
class A(object):    
    def __getattribute__(self, *args, **kwargs):    
        print "A.__getattribute__"  

class A1():    
    def __getattribute__(self, *args, **kwargs):    
        print "A1.__getattribute__"  

a1 = A1()  
a = A()  

a.test  
print "========="  
a1.test

A.__getattribute__  
=========  
Traceback (most recent call last):  
    File "t.py", line 18, in <module>  
    a1.test  
AttributeError: A1 instance has no attribute 'test'  

A是新式类，每次通过实例访问属性，都会经过__getattribute__函数,
A1不会调用__getattribute__所以出错了

​                                        抽象/实现
抽象指对现实世界问题和实体的本质表现、行为和特征**建模**,建立一个相关的子集,可以用于描绘程序
结构,从而实现这种模型。抽象不仅包括这种模型的数据属性,还定义了这些数据的接口。 对某种抽象
的实现就是对此数据及与之相关接口的现实化(realization)。现实化这个过程对于客户程序应当是透
明而且无关的。

封装/接口
封装描述了对数据/信息进行隐藏的观念,它对数据属性提供接口和访问函数。通过任何客户端直接对
数据的访问、无视接口与封装性都是背道而驰的,除非程序员允许这些操作。作为实现的一部分,客户
端根本就不需要知道在封装之后，数据属性是如何组织的。在 Python 中,所有的类属性都是公开的,
但名字可能被"混淆"了,以阻止未经授权的访问,但仅此而已,再没有其他预防措施了。这就需要在设计
时,对数据提供相应的接口,以免客户程序通过不规范的操作来存取封装的数据属性。

合成
合成扩充了对类的描述,使得多个不同的类合成为一个大的类,来解决现实问题。合成描述了一个异常
复杂的系统,比如一个类由其它类组成,更小的组件也可能是其它的类,数据属性及行为, 所有这些合在
一起,彼此是"有一个"的关系。比如,RepairShop"有一个"技工(应该至少有一个吧),还"有一个"顾客(至
少一个)。

这些组件要么通过联合关系组在一块,意思是说,对子组件的访问是允许的(对 RepairShop 来说,顾客
可能请求一个 SmogCheck,客户程序这时就是与 RepairShop 的组件进行交互),要么是聚合在一起,
封装的组件仅能通过定义好的接口来访问,对于客户程序来说是透明的。继续我的例子, 客户程序可能
会建立一个 SmogCheck 请求来代表顾客,但不能够同 RepairShop 的 SmogZone 部分进行交互,
因为 SmogZone 是由 RepairShop 内部控制的,只能通过 smogCheckCar()方法调用。Python 支
持上述两种形式的合成。

派生/继承/继承结构
派生描述了子类的创建,新类保留已存类类型中所有需要的数据和行为,但允许修改或者其它的自定义
操作,都不会修改原类的定义。继承描述了子类属性从祖先类继承这样一种方式。从前面的例子中,技
工可能比顾客多个汽车技能属性,但单独的来看,每个都"是一个"人,所以,不管 对谁而言调用 talk()都
是合法得,因为它是人的所有实例共有的。继承结构表示多"代"派生,可 以描述成一个"族谱",连续的子
类,与祖先类都有关系。

泛化/特化
泛化表示所有子类与其父类及祖先类有一样的特点,所以子类可以认为同祖先类是"是一个" 的关系,因
为一个派生对象(实例)是祖先类的一个"例子"。比如,技工"是一个"人,车"是一 个"交通工具,等等。在上
面我们间接提到的族谱图中,我们可以从子类到祖先类画一条线,表示 "是一个"的关系。特化描述所有
子类的自定义,也就是,什么属性让它与其祖先类不同。

多态
多态的概念指出了对象如何通过他们共同的属性和动作来操作及访问,而不需考虑他们具体的 类。多
态表明了动态(又名,运行时)绑定的存在,允计重载及运行时类型确定和验证。

自省/反射
自省表示给予你,程序员,某种能力来进行像"手工类型检查"的工作,它也被称为反射。这个性质展示了
某对象是如何在运行期取得自身信息的。如果传一个对象给你,你可以查出它有什么 能力,这样的功能
不是很好吗?这是一项强大的特性,在本章中,你会时常遇到。如果 Python 不支持某种形式的自省功
能,dir()和 type()内建函数,将很难正常工作。请密切关注这些调用,还有那些特殊属性,像
__dict__,__name__及__doc__。

​                                        **最简单的情况,类仅用作名称空间(namespaces) ：**
意味着你把数据保存在变量中,对他们按名称空间进行分组,使得他们处于同样的关系空间中。
所谓的关系是使用标准 Python 句点属性标识。
比如,你有一个本身没有任何属性的类,使用它仅对数据提供一个名字空间。
这样的类仅作为容器对象来共享名字空间。

例:
    class MyData(object):
        pass
    上面定义的类没有任何方法或属性。

​    一个实例,它只使用类作为名称空间容器。
​    \>>> mathObj = MyData()
​    \>>> mathObj.x = 4
​    \>>> mathObj.y = 5
​    \>>> mathObj.x + mathObj.y
​    9
​    \>>> mathObj.x * mathObj.y
​    20
当然也可以使用变量"x","y"来完成同样的事情。
但本例中,实例名字 mathObj 将 mathObj.x 和 mathObj.y 关联起来。
这就是我们所说的使用类作为名字空间容器。
mathObj.x 和 mathObj.y 是实例属性,因为它们不是类 MyData 的属性,而是实例对象(mathObj)的独有属性。
这些属性实质上是动态的:你不需要在构造器中,或其它任何地方为它们预先声明或者赋值

​                                        **静态变量和实例变量**
亦可称为静态数据（类数据属性，也叫静态字段，静态属性）。
这些数据是与它们所属的类对象绑定的,不依赖于任何类实例。
如果你熟悉 Java 或 C++,这种类型的数据相当于在一个变量声明前加上 static 关键字。
静态成员通常仅用来跟踪与类相关的值。
一般,我们会考虑用实例属性(也叫普通变量或实例变量),而不是类属性。
﻿
静态变量在内存中只保存一份
实例变量在每个对象中都保存一份

应用场景： 通过类创建对象时，如果每个对象都具有相同的字段，那么就使用静态变量
例1：类数据属性(foo): 
    \>>> class C(object):
    ...         foo = 100
    \>>> print C.foo
    100
    \>>> C.foo = C.foo + 1
    \>>> print C.foo
    101

注意,上面的代码中,看不到任何类实例的引用

例2：
    class Province:
        \# 静态变量
        country ＝ '中国'
        def __init__(self, name):
            \# 实例变量
            self.name = name
    \# 直接访问普通字段
    obj = Province('河北省')
    print obj.name
    \# 直接访问静态字段
    Province.country

​                                        要知道一个类有哪些属性,有两种方法。
​    1.最简单的是使用 dir()内建函数。
​    2.另外是通过访问类的 字典属性__dict__,这是所有类都具备的特殊属性之一。

例子:
    \>>> class MyClass(object):
    ...         'MyClass class definition'          #MyClass 类定义
    ...         myVersion = '1.1'                      # static data 静态数据
    ...         def showMyVersion(self):          # method 方法
    ...             print MyClass.myVersion 
    ...

​    根据上面定义的类,让我们使用 dir()和特殊类属性__dict__来查看一下类的属性: 
​    \>>> dir(MyClass)
​     ['__class__', '__delattr__', '__dict__', '__doc__',
​     '__getattribute__', '__hash__', '__init__', '__module__',
​     '__new__', '__reduce__', '__reduce_ex__', '__repr__',
​     '__setattr__', '__str__', '__weakref__', 'myVersion',
​     'showMyVersion']

​    \>>> MyClass.__dict__
​    <dictproxy object at 0x62090>
​    
​    \>>> print MyClass.__dict__
​    {'showMyVersion': <function showMyVersion at 0x59370>,
​    '__dict__': <attribute '__dict__' of 'MyClass' objects>,
​    'myVersion': '1.1', '__weakref__': <attribute'__weakref__' of 'MyClass' objects>,
​    '__doc__':'MyClass class definition'}

​    在新式类中,还新增加了一些属性,dir()也变得更健壮。
​    作为比较,可以看下经典类是什么样的:
​    \>>> dir(MyClass)
​    ['__doc__', '__module__', 'showMyVersion', 'myVersion']
​    
​    \>>> MyClass.__dict__
​    {'__doc__': None, 'myVersion': 1, 'showMyVersion':
​    <function showMyVersion at 950ed0>, '__module__':
​    '__main__'}

从上面可以看到,dir()返回的仅是对象的属性的一个名字列表,而__dict__返回的是一个字典, 它的键(keys)是属性
名,键值(values)是相应的属性对象的数据值。
结果还显示了 MyClass 类中两个熟悉的属性,showMyVersion 和 myVersion,以及一些新的属性。这些属
性,__doc__及__module__,是所有类都具备的特殊类属性(另外还有__dict__)。

内建的vars()函数接受类对象作为参数,返回类的__dict__属性的内容。

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**内置类访问__dict__会失败,在内建类型中,不存在这个属性:**
In [1]: a=5
In [3]: a.__dict__
\---------------------------------------------------------------------------
AttributeError                            Traceback (most recent call last)
<ipython-input-3-87da7e691200> in <module>()
----> 1 a.__dict__

AttributeError: 'int' object has no attribute '__dict__'

​                                        **特殊类属性：**
​    C.__name__       类C的名字(字符串)
​    C.__doc__          类C的文档字符串
​    C.__bases__      类C的所有父类构成的元组
​    C.__dict__          类C的属性
​    C.__module__    类C定义所在的模块
​    C.__class__        实例C对应的类(仅新式类中)

根据上面定义的类 MyClass,有如下结果:
     \>>> MyClass.__name__
    'MyClass'

​    \>>> MyClass.__doc__
​    'MyClass class definition'

​    \>>> MyClass.__bases__
​    (<type 'object'>,)

​    \>>> print MyClass.__dict__
​    {'__doc__': None, 'myVersion': 1, 'showMyVersion':
​    <function showMyVersion at 950ed0>, '__module__': '__main__'}

​    \>>> MyClass.__module__
​    '__main__'
​    
​    \>>> MyClass.__class__
​    <type 'type'>

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**__name__**
    是给定类的字符名字。
    它适用于那种只需要字符串(类对象的名字),而非类对象本身的情况。一些内建的类型也有这个属性,我们将会用
到其中之一来展示__name__字符串的益处。

​    类型对象是一个内建类型的例子,它有__name__的属性。回忆一下,type()返回被调用对象的类型。
​    我们可以使用类型对象的__name__属性来取得相应的字符串名。

如下例示:
    \>>> stype = type('What is your quest?')
    \>>> stype                       # stype is a type object stype 是一个类型对象
    <type 'string'>
    \>>> stype.__name__      # get type as a string 得到类型名(字符串表示)
    'string'

​    \>>> type(3.14159265)    # also a type object 又一个类型对象
​    <type 'float'>
​    \>>> type(3.14159265).__name__   # get type as a string 得到类型名(字符串表示) 
​    'float'

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**__doc__**
    是类的文档字符串,与函数及模块的文档字符串相似,必须紧随头行(header line) 后的字符串。
    文档字符串不能被派生类继承,也就是说派生类必须含有它们自己的文档字符串。

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**__dict__**
    包含一个字典,由类的数据属性组成。
    访问一个类属性的时候,Python 解释器将会搜索字典以得到需要的属性。
    如果在__dict__中没有找到,将会在基类的字典中进行搜索, 采用"深度优先搜索"顺序。
    基类集的搜索是按顺序的,从左到右,按其在类定义时,定义父类参数时的顺序。
    对类的修改会仅影响到此类的字典;基类的__dict__属性不会被改动的

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**__module__**
    Python 支持模块间的类继承。为更清晰地对类进行描述,1.5 版本中引入了__module__, 这样类名就完全由模
块名所限定。看一下下面的例子:
    \>>> class C(object): 
    ...         pass
    ...
    \>>> C
    <class __main__.C at 0x53f90>
    
    \>>> C.__module__
    '__main__'

​    类 C 的全名是"__main__.C",比如,source_module.class_name。
​    如果类 C 位于一个导入的模块中,如 mymod,像下面的: 
​    \>>> from mymod import C 

​    \>>> C
​    <class mymod.C at 0x53ea0> 

​    \>>> C.__module__
​    'mymod'

​    在以前的版本中,没有特殊属性__module__,很难简单定位类的位置,因为类没有使用它们的全名。
​    
\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**__class__**
​    最后,由于类型和类的统一性,当访问任何类的__class__属性时,你将发现它就是一个类型对象的实例。换句话说,
一个类已是一种类型了。因为经典类并不认同这种等价性(一个经典类是一 个类对象,一个类型是一个类型对象),对
这些对象来说,这个属性并未定义。
**__class__到底是什么意思？看例子**
​    \# cat a.py
​    \#!/usr/bin/python
​    \# -*- coding: UTF-8 -*-
​    class A:
​        def __init__(self,url):
​            self.url = url
​        def out(self):
​            return self.url

​    a = A('news.163.com')
​    print a.out()

​    b = a.__class__('www.bccn.net')
​    print b.out()

​    print A
​    print a.__class__

​    \# python a.py
​    news.163.com
​    www.bccn.net
​    __main__.A
​    __main__.A

​    **可以看出a.__class__就等效于a的类A**

​                                        实例属性 vs 类属性
类属性仅是与类相关的数据值,和实例属性不同,类属性和实例无关。这些值像静态成员那样被引用,即使在多次实例化中调用类, 它们的值都保持不变。不管如何,静态成员不会因为实例而改变它们的值,除非实例中显式改变它们的值。

类和实例都是名字空间。类是类属性的名字空间,实例则是实例属性的。
你可采用类来访问类属性,如果实例没有同名的属性的话,你也可以用实例来访问。

下面的示例中,类 C 在创建时,带一个 version 属性,这样通过类对象来访问它是很自然的了,比如,C.version。当实例 c 被创建后,对实例 c 而言,访问 c.version 会失败,不过 Python 首先会在实例中搜索名字 version,然后是类,再就是继承树中的基类。
本例中,version 在类中被找到了:
    \>>> class C(object):    # define class 定义类 
    ...         version = 1.2    # static member 静态成员 ...
    \>>> c = C() # instantiation 实例化
    \>>> C.version   # access via class 通过类来访问 
    1.2
    \>>> c.version   # access via instance 通过实例来访问 
    1.2
    \>>> C.version += 0.1 # update (only) via class 通过类(只能这样)来更新
    \>>> C.version   # class access 类访问 
    1.3
    \>>> c.version   # instance access, which 实例访问它,其值已被改变
    1.3                      # also reflected change

只有当使用类引用 version 时,才能更新它的值,像上面的 C.version 递增语句。 如果尝试在实例中设定
或更新类属性会创建一个实例属性 c.version,后者会阻止对类属性 C.versioin 的访问,因为第一个访问的就是 c.version,这样可以对实例有效地"遮蔽"类属性C.version,直到 c.version 被清除掉。

但在**类属性可变**的情况下,一切都不同了: 
    \>>> class Foo(object):
    ...         x = {2003: 'poe2'}
    ...
    \>>> foo = Foo()
    \>>> foo.x
    {2003: 'poe2'}
    \>>> foo.x[2004] = 'valid path'
    \>>> foo.x
    {2003: 'poe2', 2004: 'valid path'}
    \>>> Foo.x                                       # it works!!! 生效了
    {2003: 'poe2', 2004: 'valid path'}
    \>>> del foo.x # no shadow so cannot delete 没有遮蔽所以不能删除掉 
    Traceback (most recent call last): File "<stdin>", line 1, in ?
         del foo.x
         AttributeError: x

**类属性持久性**
静态成员,如其名所言,任凭整个实例(及其属性)的如何进展,它都不理不采(因此独立于实例)。同时,当一个实例在类属性被修改后才创建,那么更新的值就将生效。

​                                        **方法**（也算是类的属性）
我们改进类的方式之一就是给类添加功能。类的功能有一个更通俗的名字叫方法。在 Python 中,方法定义在类定义中,但只能被实例所调用。
调用一个方法的最终途径必须这样:
​    (1)定义类(和方法)
​    (2)创建一个实例
​    (3)最后一步,用这个实例调用方法。
​    
例如:
​    class MyDataWithMethod(object):     # 定义类 
​        def printFoo(self):                           # 定义方法（实例方法）
​            print 'You invoked printFoo()!'

self 参数：
    它在所有的方法（实例方法）声明中都存在。这个参数代表实例对象本身,当你用实例调用方法时,由解释器悄悄
    地传递给方法,所以,你不需要自己传递 self 进来,因为它是自动传入的。
    一般的方法会需要这个实例(self),而静态方法或类方法不会, 其中类方法需要类而不是实例。

现在我们来实例化这个类,然后调用那个方法:
\>>> myObj = MyDataWithMethod() # 创建实例 
\>>> myObj.printFoo()                        # 现在调用方法
You invoked printFoo()!

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

​                                        Python 创建实例后,在实例化过程中,调用__init__()方法,当一个类被实例化时,就可以定义额外的行为,比如,
设定初始值或者运行一些初步诊断代码—主要是在实例被创建后, 实例化调用返回这个实例之前,去执行某些
特定的任务或设置。

创建一个类(类定义)
class AddrBookEntry(object):       # 类定义 
    'address book entry class'
    def __init__(self, nm, ph):         # 定义构造器 
        self.name = nm                    # 设置 name
        self.phone = ph                    # 设置 phone
        print 'Created instance for:', self.name
    def updatePhone(self, newph): # 定义方法 
        self.phone = newph
        print 'Updated phone for:', self.name

创建实例(实例化)
    \>>> john = AddrBookEntry('John Doe', '408-555-1212') #为 John Doe 创建实例 
    \>>> jane = AddrBookEntry('Jane Doe', '650-555-1212') #为 Jane Doe 创建实例
    如果不存在默认的参数,那么传给__init__()的两个参数在实例化时是必须的。 访问实例属性
    \>>> john
    <__main__.AddrBookEntry instance at 80ee610>
    \>>> john.name
     'John Doe'
     \>>> john.phone
     '408-555-1212'
     \>>> jane.name
    'Jane Doe'
    \>>> jane.phone
    '650-555-1212'

方法调用(通过实例)
    \>>> john.updatePhone('415-555-1212') #更新 John Doe 的电话 
    \>>> john.phone
    '415-555-1212'

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**__init__()应当返回 None**
采用函数操作符调用类对象会创建一个类实例,也就是说这样一种调用过程返回的对象就是实例,下面示例可以看出:
    \>>> class MyClass(object): ... pass
    \>>> mc = MyClass()
    \>>> mc
        <__main__.MyClass instance at 95d390>
    
如果定义了构造器,它不应当返回任何对象,因为实例对象是自动在实例化调用后返回的。相 应地,__init__()就不应
当返回任何对象(应当为 None);否则,就可能出现冲突,因为只能返回实 例。试着返回非 None 的任何其它对象都会
导致 TypeError 异常:
    \>>> class MyClass:
    ...         def __init__(self):
    ...         print 'initialized'
    ...         return 1
    ...
    \>>> mc = MyClass()
             initialized
             Traceback (innermost last): File "<stdin>", line 1, in ?
             mc = MyClass()
             TypeError: __init__() should return None

### python对象销毁(垃圾回收)


同Java语言一样，Python使用了引用计数这一简单技术来追踪内存中的对象。
在Python内部记录着所有使用中的对象各有多少引用。
一个内部跟踪变量，称为一个引用计数器。
当对象被创建时，就创建了一个引用计数，当这个对象不再需要时，也就是说，这个对象的引用计数变为0 
时，它被垃圾回收。但是回收不是"立即"的， 由解释器在适当的时机，将垃圾对象占用的内存空间回收。
    a = 40      # 创建对象  <40>
    b = a        # 增加引用， <40> 的计数
    c = [b]      # 增加引用.  <40> 的计数

​    del a          # 减少引用 <40> 的计数
​    b = 100     # 减少引用 <40> 的计数
​    c[0] = -1    # 减少引用 <40> 的计数 
垃圾回收机制不仅针对引用计数为0的对象，同样也可以处理循环引用的情况。
循环引用指的是，两个对象相互引用，但是没有其他变量引用他们。这种情况下，仅使用引用计数是不够的。
Python 的垃圾收集器实际上是一个引用计数器和一个循环垃圾收集器。作为引用计数的补充，垃圾收集器也
会留心被分配的总量很大（及未通过引用计数销毁的那些）的对象。 在这种情况下， 解释器会暂停下来， 试
图清理所有未引用的循环。  

例1：
析构函数 __del__ ，**__del__在****对象****销毁的时候被调用，当对象不再被使用时，__del__方法运行：** 
​    \#!/usr/bin/python
​    \# -*- coding: UTF-8 -*-
​    class Point:
​       def __init__( self, x=0, y=0):
​          self.x = x
​          self.y = y
​       def __del__(self):
​          class_name = self.__class__.__name__
​          print class_name, "销毁"

​    pt1 = Point()
​    pt2 = pt1
​    pt3 = pt1
​    print id(pt1), id(pt2), id(pt3) # 打印对象的id
​    del pt1
​    del pt2
​    del pt3

以上实例运行结果如下：
    3083401324 3083401324 3083401324
    Point 销毁 

例2：
\#!/usr/bin/env python
\#coding=utf8
class InstCt(object):
    count = 0                         # count is class attr count 是一个类属性
    def __init__(self):             # increment count 增加 count 
        InstCt.count += 1
    def __del__(self):              # decrement count 减少 count 
        InstCt.count -= 1
    def howMany(self):          # return count 返回 count 
        return InstCt.count

a = InstCt()
b = InstCt()
print b.howMany()
print a.howMany()
del b
print a.howMany()
del a
print InstCt.count

​                                        **绑定(绑定及非绑定方法)**
方法仅仅是类内部定义的函数(这意味着方法是类属性而不是实例属性)
为与 OOP 惯例保持一致,Python 严格要求,没有实例,方法是不能被调用的。这种限制即Python所描述的绑定概念(binding),当存在一个实例时,方法才被认为是绑定到那个实例了。没有实例时方法就是未绑定的。任何一个方法
定义中的第一个参数都是变量 self,它表示调用此方法的实例对象。

实例：
In [10]: class A(object):
    a=8
    def func(self):
        print "这是方法，不能在没有实例的时候调用"
   ....:         

In [11]: A.func()
\---------------------------------------------------------------------------
TypeError                                 Traceback (most recent call last)
<ipython-input-11-733ae5a20fb4> in <module>()
----> 1 A.func()

TypeError: unbound method func() must be called with A instance as first argument (got nothing 
instead)

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**self 是什么?**
self 变量用于在类实例方法中引用方法所绑定的实例。因为方法的实例在任何方法调用中总是 作为第一个参数传
递的,self 被选中用来代表实例,你必须在方法声明中放上 self

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**调用绑定方法**
方法,不管绑定与否,都是由相同的代码组成的。唯一的不同在于是否存在一个实例可以调用此方法。在很多情况下,
程序员调用的都是一个绑定的方法。假定现在有一个 MyClass 类和此类的一个实例 mc,而你想调用 
MyClass.foo()方法。因为已经有一个实例,你只需要调用 mc.foo()就可 以。记得 self 在每一个方法声明中都是作
为第一个参数传递的。当在实例中调用一个绑定的方法 时,self 不需要明确地传入了。这算是"必须声明 self 作
为第一个参数"对你的报酬。当你还没有一个实例并且需要调用一个非绑定方法的时候你必须传递 self 参数。

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**调用非绑定方法**
调用非绑定方法并不经常用到。
需要调用一个还没有任何实例的类中的方法的一个主要的场景是:
    你在派生一个子类,而且你要覆盖父类的方法,这时你需要调用那个父类中想要覆盖掉的构造方法。

例子:
class EmplAddrBookEntry(AddrBookEntry):
'Employee Address Book Entry class' # 员工地址记录条目
     def __init__(self, nm, ph, em):
         AddrBookEntry.__init__(self, nm, ph)
         self.empid = id
         self.email = em
         
EmplAddrBookEntry 是 AddrBookEntry 的子类,我们重载了构造器__init__()。我们想尽可能多地重用代码, 而
不是去从父类构造器中剪切,粘贴代码。这样做还可以避免 BUG 传播,因为任何修复都可以传递给子类。这正是我
们想要的 --- 没有必要一行一行地复制代码。只需要能够调用父类的构造器即可,但该怎么做呢?
我们在运行时没有 AddrBookEntry 的实例。那么我们有什么呢?我们有一个 EmplAddrBookEntry 的实例,它与 
AddrBookEntry 是那样地相似,我们难道不能用它代替呢?当然可以!
当一个 EmplAddrBookEntry 被实例化,并且调用 __init__() 时,其与 AddrBookEntry 的实例只有很少的差别,主
要是因为我们还没有机会来自定义我们的 EmplAddrBookEntry 实例,以使它与 AddrBookEntry 不同。
这是调用非绑定方法的最佳地方了。我们将在子类构造器中调用父类的构造器并且明确地传递 (父类)构造器所需要
的 self 参数(因为我们没有一个父类的实例)。子类中 __init__() 的第一行就是对父类__init__()的调用。我们通过
父类名来调用它,并且传递给它 self 和其他所需要的参数。一旦调用返回,我们就能定义那些与父类不同的仅存在我
们的(子)类中的(实例)定制。

​                                        **静态方法：**
​    仅是类中的函数(不需要实例调用)
​    ﻿普通的方法，可以在实例化后直接调用，并且在方法里可以通过self.调用实例变量或类变量，但静态方法是不可以访问实例变量或类变量的，一个不能访问实例变量和类变量的方法，其实相当于跟类本身已经没什么关系了，它与类唯一的关联就是需要通过类名来调用这个方法
**类方法：**
​    通常的方法需要一个实例(self)作为第一个参数,并且对于(绑定的)方法调用来 说,self 是自动传递给这个方法的。
​    而对于类方法而言,需要类而不是实例作为第一个参数,它是由解释器传给方法。类不需要特别地命名, 类似 self,不过很多人使用 cls 作为变量名字。（不需要实例调用）

**staticmethod()和 classmethod()内建函数**
使用对应的内建函数转换它们为相应的类型,并且重新赋值给了相同的变量名。如果没有调用这两个函数,二者都会在 Python 编译器中产生错误,显示需要带 self 的常规方法声明。现在, 我们可以通过类或者实例调用这些函数这没什么不同:
\#!/usr/bin/python
\# -*- coding: UTF-8 -*-
class TestStaticMethod:
        def foo():
            print 'calling static method foo()'
        foo = staticmethod(foo)   #把函数转换成静态方法
class TestClassMethod:
        def foo(cls):             **#这里的cls（class缩写）可以是任何字符，只是约定写为cls**
            print 'calling class method foo()'
            print 'foo() is part of class:', cls.__name__
        foo = classmethod(foo)    #把函数转换成类方法

TestStaticMethod.foo()    **#静态方法可以直接被类调用，不用实例（普通方法没有实例不能被调用）**
a=TestStaticMethod()
a.foo()
TestClassMethod.foo()     **#类方法可以直接被类调用，不用实例（普通方法没有实例不能被调用）**
b=TestClassMethod()
b.foo()

**函数修饰符**
你可以用它把一个函数应用到另个函数对象上, 而且新函数对象依然绑定在原来的变量。我们正是需要它来整理语法。通过使用 decorators,我们可以避免像上面那样的重新赋值:
\#!/usr/bin/python
class TestStaticMethod:
        @staticmethod
        def foo():
            print 'calling static method foo()'
class TestClassMethod:
        @classmethod
        def foo(cls):
            print 'calling class method foo()'
            print 'foo() is part of class:', cls.__name__

TestStaticMethod.foo()    **#静态方法可以直接被类调用，不用实例（普通方法没有实例不能被调用）**
a=TestStaticMethod()
a.foo()
TestClassMethod.foo()     **#类方法可以直接被类调用，不用实例（普通方法没有实例不能被调用）**
b=TestClassMethod()
b.foo()

​                                        如果说类是一种数据结构定义类型,那么实例则声明了一个这种类型的变量。换言之,实例是有生命的类。就像设计完一张蓝图后,就是设法让它成为现实。实例是那些主要用在运行期时的对象,类被实例化得到实例,该实例的类型就是这个被实例化的类

初始化:
    通过调用类对象来创建实例
    很多其它的 OO 语言都提供 new 关键字,通过 new 可以创建类的实例。Python 的方式更加简单。 一旦定义了一个类,创建实例比调用一个函数还容易------不费吹灰之力。

​    实例化的实现,可以使用函数操作符,如下示:
​        \>>> class MyClass(object):   # define class 定义类 
​        ...         pass
​        \>>> mc = MyClass()              # instantiate class 初始化类

​    可以看到,仅调用("calling")类:MyClass(),就创建了类 MyClass 的实例 mc。返回的对象是你所调用类的一个实
例。当使用函数记法来调用("call")一个类时,解释器就会实例化该对象

实例属性
实例仅拥有数据属性(方法严格来说是类属性),后者只是与某个类的实例相关联的数据值,并且可以通过句点属性标
识法来访问。这些值独立于其它实例或类。当一个实例被释放后,它的属性同时也被清除了。

"实例化"实例属性(或创建一个更好的构造器)
设置实例的属性可以在实例创建后任意时间进行,也可以在能够访问实例的代码中进行。构造器__init()__是设置这
些属性的关键点之一。
能够在"运行时"创建实例属性,是 Python 类的优秀特性之一,从 C++或 Java 转过来的人会被 小小的震惊一下,因
为 C++或 Java 中所有属性在使用前都必须明确定义/声明。
Python 不仅是动态类型,而且在运行时,允许这些对象属性的动态创建。这种特性让人爱不释手。当然,我们必须提
醒读者,创建这样的属性时,必须谨慎。
一个缺陷是,属性在条件语句中创建,如果该条件语句块并未被执行,属性也就不存在,而你在后面的代码中试着去访
问这些属性,就会有错误发生。故事的精髓是告诉我们,Python 让你体验 从未用过的特性,但如果你使用它了,你还
是要小心为好。

在构造器中首先设置实例属性
构造器是最早可以设置实例属性的地方,因为__init__()是实例创建后第一个被调用的方法。 再没有比这更早的可以
设置实例属性的机会了。一旦__init__()执行完毕,返回实例对象,即完成了实例化过程。

​                                        **组合**
​    一个类被定义后,目标就是要把它当成一个模块来使用,并把这些对象嵌入到你的代码中去, 同其它数据类型及逻辑执行流混合使用。有两种方法可以在你的代码中利用类。
​    第一种是组合 (composition)。就是让不同的类混合并加入到其它类中,来增加功能和代码重用性。你可以在一个大点的类中创建你自已的类的实例,实现一些其它属性和方法来增强对原来的类对象。
​    另一种方法是通过派生

​    举例来说,让我们想象创建的地址本类的加强性设计。如果在设计的过程中, 为 names,addresses 等等创建了单独的类。那么最后我们可能想把这些工作集成到 AddrBookEntry 类中去,而不是重新设计每一个需要的类。这样就节省了时间和精力,而且最后的结果是容易维护的代码 --- 一块代码中的 bugs 被修正,将反映到整个应用中。
​    这样的类可能包含一个 Name 实例,以及其它的像 StreetAddress, Phone(home, work, telefacsimile, 
pager, mobile, 等等),Email (home, work, 等等。),还可能需要一些 Date实例
(birthday,wedding,anniversary,等等)。

一个简单的例子:
class NewAddrBookEntry(object): # class definition 类定义 
    'new address book entry class'
    def __init__(self, nm, ph):             # define constructor 定义构造器
        self.name = Name(nm)                # create Name instance 创建 Name 实例 
        self.phone = Phone(ph)                # create Phone instance 创建 Phone 实例 
        print 'Created instance for:', self.name

NewAddrBookEntry 类由它自身和其它类组合而成。这就在一个类和其它组成类之间定义了一种 "has-a / 有一个"的关系。比如,我们的 NewAddrBookEntry 类"有一个" Name 类实例和一个 Phone 实例。
创建复合对象就可以实现这些附加的功能,并且很有意义,因为这些类都不相同。每一个类管理它们自己的名字空间和行为。不过当对象之间有更接近的关系时,派生的概念可能对你的应用程序来说更有意义,特别是当你需要一些相似的对象,但却有少许不同功能的时候。

​                                        **概念：**
当类之间有显著的不同,并且(较小的类)是较大的类所需要的组件时,组合表现得很好,但当你设计"相同的类但有
一些不同的功能"时,派生就是一个更加合理的选择了。

OOP 的更强大方面之一是能够使用一个已经定义好的类,扩展它或者对其进行修改,而不会影响 系统中使用现存
类的其它代码片段。OOD 允许类特征在子孙类或子类中进行继承。这些子类从基类(或 称祖先类,超类)继承它
们的核心属性。而且,这些派生可能会扩展到多代。在一个层次的派生关 系中的相关类(或者是在类树图中垂直
相邻)是父类和子类关系。从同一个父类派生出来的这些类 (或者是在类树图中水平相邻)是同胞关系。父类和所
有高层类都被认为是祖先

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**创建子类**
创建子类的语法看起来与普通(新式)类没有区别,一个类名,后跟一个或多个需要从其中派生的父类:
    class SubClassName (ParentClass1[, ParentClass2, ...]):
        'optional class documentation string'
        class_suite

如果你的类没有从任何祖先类派生,可以使用 object 作为父类的名字。经典类的声明唯一不同之处在于其没有
从祖先类派生---此时,没有圆括号:
    class ClassicClassWithoutSuperclasses:
        pass

例子: 
class Parent(object):                   # define parent class 定义父类
    def parentMethod(self):
         print 'calling parent method'

class Child(Parent):                    # define child class 定义子类 
    def childMethod(self):
        print 'calling child method'

\>>> p = Parent()            # instance of parent 父类的实例 
\>>> p.parentMethod()
calling parent method
\>>> c = Child()              # instance of child 子类的实例 
\>>> c.childMethod()     # child calls its method 子类调用它的方法
calling child method
\>>> c.parentMethod()  # calls parent's method 调用父类的方法
calling parent method

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**继承**
继承描述了基类的属性如何"遗传"给派生类。一个子类可以继承它的基类的任何属性,不管是数据属性还是方法。
例子：
P 是一个没有属性的简单类。C 从 P 继承而来(因此是它的子类),也没有属性:
    class P(object):                 # parent class 父类 pass
    class C(P):                         # child class 子类
        pass

​    \>>> c = C()                     # instantiate child 实例化子类
​    \>>> c.__class__               # child "is a" parent 子类"是一个"父类 
​    <class '__main__.C'>
​    \>>> C.__bases__             # child's parent class(es) 子类的父类 
​    (<class '__main__.P'>,)

因为 P 没有属性,C 没有继承到什么。下面我们给 P 添加一些属性: 
    class P:              # parent class 父类
        'P class'
        def __init__(self):
            print 'created an instance of', self.__class__.__name__

​    class C(P):         # child class 子类 
​         pass

现在所创建的 P 有文档字符串(__doc__)和构造器,当我们实例化 P 时它被执行,如下面的交互会话所示:
    \>>> p = P()                     # parent instance 父类实例
    created an instance of P
    \>>> p.__class__             # class that created us 显示p所属的类名 
    <class '__main__.P'>
    \>>> P.__bases__            # parent's parent class(es) 父类的父类 
    (<type 'object'>,)
    \>>> P.__doc__               # parent's doc string 父类的文档字符串 
    'P class'

我们现在来实例化 C,展示 __init__()(构造)方法在执行过程中是如何继承的:
    \>>> c = C()                    # child instance 子类实例
    created an instance of C
    \>>> c.__class__              # class that created us 显示c所属的类名 
    <class '__main__.C'>
    \>>> C.__bases__            # child's parent class(es) 子类的父类 
    (<class '__main__.P'>,)
    \>>> C.__doc__                # child's doc string 子类的文档字符串

C 没有声明__init__()方法,然而在类 C 的实例 c 被创建时,还是会有输出信息。原因在于 C 继承了 P 的
__init__()。
__bases__元组列出了其父类 P。
需要注意的是文档字符串对类,函数/方法, 还有模块来说都是唯一的,所以特殊属性__doc__不会从基类中继承过
来。

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**__bases__类属性**
__bases__类属性,对任何(子)类,它是一个包含其父类 (parent)的集合的元组。那些没有父类的类,它们的
__bases__属性为空。

下面我们看一下如何使用__bases__:
    \>>> class A(object): pass           # define class A   定义类 A 
    \>>> class B(A): pass                   # subclass of A    A 的子类 
    \>>> class C(B): pass                   # subclass of B (and indirectly, A)  B 的子类(A 的间接子类) 
    \>>> class D(A, B): pass              # subclass of A and B A,B 的子类 
    \>>> A.__bases__
    (<type 'object'>,)
    \>>> C.__bases__
    (<class __main__.B at 8120c90>,)
    \>>> D.__bases__
    (<class __main__.A at 811fc90>, <class __main__.B at 8120c90>)

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**通过继承覆盖(Overriding)方法**
我们在 P 中再写一个函数,然后在其子类中对它进行覆盖。 
    class P(object):
         def foo(self):
             print 'Hi, I am P-foo()'
    \>>> p = P()
    \>>> p.foo()
        Hi, I am P-foo()

现在来创建子类 C,从父类 P 派生: 
    class C(P):
         def foo(self):
             print 'Hi, I am C-foo()'
    \>>> c = C()
    \>>> c.foo()
    Hi, I am C-foo()
尽管 C 继承了 P 的 foo()方法,但因为 C 定义了它自已的 foo()方法,所以 P 中的 foo() 方法被覆盖。覆盖方法
的原因之一是,你的子类可能需要这个方法具有特定或不同的功能。

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**调用被覆盖的基类方法：**
调用一个未绑定的基类方法,需要明确给出子类的实例,例如:
     \>>> P.foo(c)
     Hi, I am P-foo()
注意,上面已经有了一个 P 的实例 p,但上面的这个例子并没有用它。不需要 P 的实例调用 P 的方法,因为已经有
一个 P 的子类的实例 c 可用。典型情况下,你不会以这种方式调用父类方法,你会在子类的重写方法里显式地调
用基类方法。
    class C(P):
        def foo(self):
            P.foo(self)
            print 'Hi, I am C-foo()'
注意,在这个(未绑定)方法调用中我们显式地传递了 self. 一个更好的办法是使用 super() 内建方法:
    class C(P):
        def foo(self):
            super(C, self).foo()
            print 'Hi, I am C-foo()'
super()不但能找到基类方法,而且还为我们传进 self,这样我们就不需要做这些事了。现在我们只要调用子类的
方法,它会帮你完成一切:
    \>>> c = C()
    \>>> c.foo()
    Hi, I am P-foo() Hi, I am C-foo()

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**重写__init__不会自动调用基类的__init__**
类似于上面的覆盖非特殊方法,当从一个带构造器 __init()__的类派生,如果你不去覆盖 __init__(),它将会被继承
并自动调用。但如果你在子类中覆盖了__init__(),子类被实例化时, 基类的__init__()就不会被自动调用。这可
能会让了解 JAVA 的朋友感到吃惊。
    class P(object):
        def __init__(self):
            print "calling P's constructor"
    class C(P):
        def __init__(self):
            print "calling C's constructor"
    \>>> c = C()
    calling C's constructor

如果你还想调用基类的 __init__(),你需要明确指出,使用一个子类的实例去调用基类(未绑定)方法。相应地更新
类 C,会出现下面预期的执行结果:
     class C(P):
        def __init__(self):
            P.__init__(self)
            print "calling C's constructor"

​     \>>> c = C()
​    calling P's constructor
​    calling C's constructor 
上边的例子中,子类的__init__()方法首先调用了基类的的__init__()方法。这是相当普遍(不是强制)的做法,用来
设置初始化基类,然后可以执行子类内部的设置。这个规则之所以有意义的原因是,你希望被继承的类的对象在
子类构造器运行前能够很好地被初始化或作好准备工作,因为它 (子类)可能需要或设置继承属性。

**内建函数super()**
对 C++熟悉的童鞋,可能会在派生类构造器声明时,通过在声明后面加上冒号和所要调用的所有基类构造器这种
形式来调用基类构造器。而在 JAVA 中,不管程序员如何处理,子类构造器都会去调用基类的的构造器。

Python 使用基类名来调用类方法,对应在 JAVA 中,是用关键字 super 来实现的,这就是 super() 内建函数引入
到 Python 中的原因,这样你就可以"依葫芦画瓢"了:
    class C(P):
        def __init__(self):
            super(C, self).__init__()
            print "calling C's constructor"

使用 super()的漂亮之处在于,你不需要明确给出任何基类名字..."跑腿事儿",它帮你干了! 使用 super()的重点,是你不需要明确提供父类。这意味着如果你改变了类继承关系,你只需要改一行代码(class 语句本身)而不必在大量代码中去查找所有被修改的那个类的名字。

​                                        每个子类最好定义它自己的构造器,不然,基类的构造器会被调用。
如果子类重写基类的构造器,基类的构造器就不会被自动调用了--这样,基类的构造器就必须显式写出才会被执行,像
我们上面那样,用 AddrBookEntry.__init__()设置名字和电话号码。我们的子类在构造器后面几行还设置了另外两
个实例属性:员工 ID 和 E-mail 地址。
​    class EmplAddrBookEntry(AddrBookEntry): 
​    'Employee Address Book Entry class'#员工地址本类
​            def __init__(self, nm, ph, id, em):
​                AddrBookEntry.__init__(self, nm, ph)
​                self.empid = id
​                self.email = em
​            def updateEmail(self, newem):
​                self.email = newem
​                print 'Updated e-mail address for:', self.name

​    \>>> john = EmplAddrBookEntry('John Doe', '408-555-1212',42, 'john@spam.doe') 
​    Created instance for: John Doe                 #给 John Doe 创建实例
​    \>>> john
​    <__main__.EmplAddrBookEntry object at 0x62030>

​    \>>> john.name
​    'John Doe'
​    \>>> john.phone
​    '408-555-1212'
​    \>>> john.email
​    'john@spam.doe'
​    \>>> john.updatePhone('415-555-1212') 
​    Updated phone for: John Doe
​    \>>> john.phone
​    '415-555-1212'
​    \>>> john.updateEmail('john@doe.spam') 
​    Updated e-mail address for: John Doe
​    \>>> john.email
​    'john@doe.spam'

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

命名类、属性和方法
类名通常由大写字母打头。这是标准惯例,可以帮助你识别类,特别是在实例化过程中(有时看起来像函数调用)。
数据属性(变量或常量)应当是数据值的名字,方法名应当指出对应对象或值的行为。

另一种表达方式是:数据值应该使用名词作为名字,方法使用谓词(动词加对象)。数据项是操作的对象、方法应当表
明程序员想要在对象进行什么操作。
在上面我们定义的类中,遵循了这样的方针,数据值像"name","phone"和"email",行为如"updatePhone", 
"updateEmail"。这就是常说的"混合记法(mixedCase)"或"骆驼记法(camelCase)"。Python 规范推荐使用骆驼
记法的下划线方式,比如,"update_phone","update_email"。
类也要细致命名, 像"AddrBookEntry","RepairShop"等等就是很好的名字。

​                                        调用父类的构造函数
子类（派生类）并不会自动调用父类（基类）的init方法，例如：

class Foo(object):
    def __init__(self):
        self.val = 1
class Foo2(Foo):
    def __init__(self):
        print self.val
if __name__ == '__main__':
    foo2 = Foo2()

运行时报错。

调用父类的init方法有两种：
第一种：
class Foo(object):
    def __init__(self):
        self.val = 1
class Foo2(Foo):
    def __init__(self):
        Foo.__init__(self)
        print self.val
if __name__ == '__main__':
    foo2 = Foo2()

第二种：
class Foo(object):
    def __init__(self):
        self.val = 1
class Foo2(Foo):
    def __init__(self):
        super(Foo2,self).__init__()
        print self.val
if __name__ == '__main__':
    foo2 = Foo2()
\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**super**
super用来执行父类中的函数，例如：
class Foo(object):
    def hi(self):
        print 'hi,Foo'
class Foo2(Foo):
    def hi(self):
        super(Foo2, self).hi()
if __name__ == '__main__':
    foo2 = Foo2()
    foo2.hi()

运行结果：
hi,Foo

注意，Foo类必须继承某个类（并且这个继承链开始于object类），否则会报错。如果改成下面的形式：
class Foo:
    def hi(self):
        print 'hi,Foo'
class Foo2(Foo):
    def hi(self):
        super(Foo2, self).hi()
if __name__ == '__main__':
    foo2 = Foo2()
    foo2.hi()

运行时报错如下：
......
TypeError: must be type, not classobj

​                                        用来对float做四舍五入操作：
In [4]: a=2.888888

In [5]: round(a,2)
Out[5]: 2.89

In [6]: round(a,3)
Out[6]: 2.889

In [7]: round(a,1)
Out[7]: 2.9

​                                        **标准类型派生**
经典类中,一个最大的问题是,不能对标准类型进行子类化。幸运的是,在 2.2 以后的版本中, 随着类型(types)和
类(class)的统一和新式类的引入, 这一点已经被修正。
下面,介绍两个子类化 Python 类型的相关例子,其中一个是可变类型,另一个是不可变类型。

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**不可变类型的例子**
假定你想在金融应用中,应用一个处理浮点数的子类。每次你得到一个贷币值(浮点数), 你都需要通过四舍五入,
变为带两位小数位的数值。你的类开始可以这样写:
     class RoundFloat(float):
        def __new__(cls, val):
            return float.__new__(cls, round(val, 2))
我们覆盖了__new__()特殊方法来定制我们的对象,使之和标准 Python 浮点数(float)有一些区别:我们使用 
round()内建函数对原浮点数进行舍入操作,然后实例化我们的 float,RoundFloat。 我们是通过调用父类的构
造器来创建真实的对象的,float.__new__()。注意,所有的__new()__方法都是类方法,我们要显式传入类传为
第一个参数,这类似于常见的方法如__init__()中需要的 self。
现在的例子还非常简单,比如,我们知道有一个 float,我们仅仅是从一种类型中派生而来等等. 通常情况下,最好是
使用 super()内建函数去捕获对应的父类以调用它的__new()__方法,下面,对它进行这方面的修改:
    class RoundFloat(float):
        def __new__(cls, val):
        return super(RoundFloat, cls).__new__(cls, round(val, 2)) 
这个例子还远不够完整,所以,请留意本章我们将使它有更好的表现。下面是一些样例输出:
    \>>> RoundFloat(1.5955)
    1.6
    \>>> RoundFloat(1.5945)
    1.59
    \>>> RoundFloat(-1.9955)
    -2.0
\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**可变类型的例子**
子类化一个可变类型与此类似,你可能不需要使用__new__() (或甚至__init__()),因为通常设置不多。一般情况
下,你所继承到的类型的默认行为就是你想要的。
下例中,我们简单地创建一 个新的字典类型,它的 keys()方法会自动排序结果:
    class SortedKeyDict(dict):
        def keys(self):
            return sorted(super( SortedKeyDict, self).keys())

回忆一下,字典(dictionary)可以由 dict(),dict(mapping),dict(sequence_of_2_tuples), 或者 dict
(**kwargs)来创建,看看下面使用新类的例子:

​     d = SortedKeyDict((('zheng-cai', 67), ('hui-jun', 68),('xin-yi', 2)))
​     print 'By iterator:'.ljust(12), [key for key in d]
​     print 'By keys():'.ljust(12), d.keys()

把上面的代码全部加到一个脚本中,然后运行,可以得到下面的输出:
By iterator: ['zheng-cai', 'xin-yi', 'hui-jun'] By keys(): ['xin-yi', 'hui-jun', 'zheng-cai']
在上例中,通过 keys 迭代过程是以散列顺序的形式,而使用我们(重写的)keys()方法则将 keys 变为字母排序方
式了。
一定要谨慎,而且要意识到你正在干什么。如果你说,“"的方法调用 super()过于复杂",取而代之的是,你更喜欢 
keys()简简单单(也容易理解)....,像这样:
    def keys(self):
        return sorted(self.keys())

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

多重继承
同 C++一样,Python 允许子类继承多个基类。这种特性就是通常所说的多重继承。概念容易, 但最难的工作
是,如何正确找到没有在当前(子)类定义的属性。当使用多重继承时,有两个不同的方面要记住。首先,还是要找
到合适的属性。另一个就是当你重写方法时,如何调用对应父类方法以"发挥他们的作用",同时,在子类中处理好
自己的义务。我们将讨论两个方面,但侧重后者, 讨论方法解析顺序。

方法解释顺序(MRO)
在 Python 2.2 以前的版本中,算法非常简单:深度优先,从左至右进行搜索,取得在子类中使用的属性。其它 
Python 算法只是覆盖被找到的名字,多重继承则取找到的第一个名字。
由于类,类型和内建类型的子类,都经过全新改造, 有了新的结构,这种算法不再可行. 这样一种新的 MRO 算法被
开发出来,在 2.2 版本中初次登场,是一个好的尝试,但有一个缺陷(看下面的核心笔记)。这在 2.3 版本中立即被
修改,也就是今天还在使用的版本。

精确顺序解释很复杂,超出了本文的范畴,但你可以去阅读本节后面的参考书目提到的有关内
容。这里提一下,新的查询方法是采用广度优先,而不是深度优先。

Python 2.2 使用一种唯一但不完善的 MRO
Python 2.2 是首个使用新式 MRO 的版本,它必须取代经典类中的算法,原因在上面已谈到过。
在 2.2 版本中,算法基本思想是根据每个祖先类的继承结构,编译出一张列表,包括搜索到的类, 按策略删除重复
的。然而,在 Python 核心开发人员邮件列表中,有人指出,在维护单调性方面失败过(顺序保存),必须使用新的 
C3 算法替换,也就是从 2.3 版开始使用的新算法。
下面的示例,展示经典类和新式类中,方法解释顺序有什么不同。
简单属性查找示例
下面这个例子将对两种类的方案不同处做一展示。脚本由一组父类,一组子类,还有一个子孙类组成。
class P1: #(object): # parent class 1 父类 1 def foo(self):
         print 'called P1-foo()'
class P2: #(object): # parent class 2 父类 2 def foo(self):
        print 'called P2-foo()'
    def bar(self):
     print 'called P2-bar()'
class C1(P1, P2): # child 1 der. from P1, P2 #子类 1,从 P1,P2 派生 pass
class C2(P1, P2): # child 2 der. from P1, P2 #子类 2,从 P1,P2 派生 def bar(self):
         print 'called C2-bar()'
class GC(C1, C2): # define grandchild class #定义子孙类 pass # derived from C1 and C2 #从 C1,C2 派生
[![file://C:\Users\86186\AppData\Local\Temp\.2IK5Q0\1.png](assets/1-1600310837640.png)]()  
父类,子类及子孙类的关系图,还有它们各自定义的方法

在图 13-2 中,我们看到父类,子类及子孙类的关系。P1 中定义了 foo(),P2 定义了 foo()和 bar(), C2 定义了 
bar()。下面举例说明一下经典类和新式类的行为。
经典类
首先来使用经典类。通过在交互式解释器中执行上面的声明,我们可以验证经典类使用的解释顺序,深度优先,从
左至右:
\>>> gc = GC()
\>>> gc.foo() # GC ==> C1 ==> P1
called P1-foo()
\>>> gc.bar() # GC ==> C1 ==> P1 ==> P2 called P2-bar()
当调用 foo()时,它首先在当前类(GC)中查找。如果没找到,就向上查找最亲的父类,C1。查找未遂,就继续沿树上
访到父类 P1,foo()被找到。
同样,对 bar()来说,它通过搜索 GC,C1,P1 然后在 P2 中找到。因为使用这种解释顺序的缘故, C2.bar()根本就
不会被搜索了。
现在,你可能在想,"我更愿意调用 C2 的 bar()方法,因为它在继承树上和我更亲近些,这样才会更合适。"在这种
情况下,你当然还可以使用它,但你必须调用它的合法的全名,采用典型的非绑定方式去调用,并且提供一个合法的
实例:
    \>>> C2.bar(gc)
    called C2-bar()
新式类
取消类 P1 和类 P2 声明中的对(object)的注释,重新执行一下。新式方法的查询有一些不同:
\>>> gc = GC()
\>>> gc.foo() # GC ==> C1 ==> C2 ==> P1 called P1-foo()
\>>> gc.bar() # GC ==> C1 ==> C2 called C2-bar()
与沿着继承树一步一步上溯不同,它首先查找同胞兄弟,采用一种广度优先的方式。当查找 foo(),它检查 GC,然
后是 C1 和 C2,然后在 P1 中找到。如果 P1 中没有,查找将会到达 P2。foo() 的底线是,包括经典类和新式类都
会在 P1 中找到它,然而它们虽然是同归,但殊途!
然而,bar()的结果是不同的。它搜索 GC 和 C1,紧接着在 C2 中找到了。这样,就不会再继续搜 索到祖父 P1 和 
P2。这种情况下,新的解释方式更适合那种要求查找 GC 更亲近的 bar()的方案。当然,如果你还需要调用上一
级,只要按前述方法,使用非绑定的方式去做,即可。
    \>>> P2.bar(gc)
    called P2-bar()
新式类也有一个__mro__属性,告诉你查找顺序是怎样的:
    \>>> GC.__mro__
    (<class '__main__.GC'>, <class '__main__.C1'>, <class
    '__main__.C2'>, <class '__main__.P1'>, <class
    '__main__.P2'>, <type 'object'>)
菱形效应为难 MRO
经典类方法解释不会带来很多问题。它很容易解释,并理解。大部分类都是单继承的,多重继承只限用在对两个
完全不相关的类进行联合。这就是术语 mixin 类(或者"mix-ins")的由来。
为什么经典类 MRO 会失败
在版本 2.2 中,类型与类的统一,带来了一个新的"问题",波及所有从 object(所有类型的祖先类)派生出来的(根)
类,一个简单的继承结构变成了一个菱形。从 Guido van Rossum 的文章中得到下面的灵感,打个比方,你有经
典类 B 和 C,C 覆盖了构造器,B 没有,D 从 B 和 C 继承而来:
class B: pass
    class C:
     def __init__(self):
         print "the default constructor"
    class D(B, C):
        pass
当我们实例化 D,得到:
    \>>> d = D()
    the default constructor
图 13.3 为 B,C 和 D 的类继承结构,现在把代码改为采用新式类的方式,问题也就产生了:
    class B(object):
     pass
    class C(object):
     def __init__(self):
         print "the default constructor"
  [![file://C:\Users\86186\AppData\Local\Temp\.2IK5Q0\2.png](assets/2-1600310837640.png)]()      

  图 13.3 继承的问题是由于在新式类中,需要出现基类,这样就在继承结构中,形成了一个菱形。D 的实例上溯时,不应当错过 C,但不能两次上溯到 A(因为 B 和 C 都从 A 派生)。去读读 Guido van Rossum 的文章中有关"协作方法"的部分,可以得到更深地理解。
代码中仅仅是在两个类声明中加入了(object),对吗?没错,但从图中,你可以看出,继承结构已变成了一个菱形;真正的问题就存在于 MRO 了。如果我们使用经典类的 MRO,当实例化 D 时,不再得到 C.__init__()之结果.....而是得到 object.__init__()!这就是为什么 MRO 需要修改的真正原因。
尽管我们看到了,在上面的例子中,类 GC 的属性查找路径被改变了,但你不需要担心会有大量的代码崩溃。经典类将沿用老式 MRO,而新式类将使用它自己的 MRO。还有,如果你不需要用到新式类中的所有特性,可以继续使用经典类进行开发,不会有问题的。

总结
经典类,使用深度优先算法。因为新式类继承自 object,新的菱形类继承结构出现,问题也就接着而来了,所以必
须新建一个 MRO。
你可以在下面的链接中读在更多有关新式类、MRO 的文章:
Guido van Rossum 的有关类型和类统一的文章: http://www.python.org/download/releases/2.2.3/descrintro
PEP 252:使类型看起来更像类 http://www.python.org/doc/peps/pep-0252
"Python 2.2 新亮点" 文档 http://www.python.org/doc/2.2.3/whatsnew
论文:Python 2.3 方法解释顺序 http://python.org/download/releases/2.3/mro/ 
                                                 Python的可变类型与不可变类型

Python的每个对象都分为可变和不可变，主要的核心类型中，数字、字符串、元组是不可变的，列表、字典是
可变的。
对不可变类型的变量重新赋值，实际上是重新创建一个不可变类型的对象，并将原来的变量重新指向新创建的
对象（如果没有其他变量引用原有对象的话（即引用计数为0），原有对象就会被回收）。

不可变类型以int类型为例:实际上 i += 1 并不是真的在原有的int对象上+1，而是重新创建一个value为6的int
对象，i引用自这个新的对象。
\>>> i = 5
\>>> i += 1
\>>> i
6
通过id函数查看变量i的内存地址进行验证（使用hex(id(i)) 可以查看16进制的内存地址）

\>>> i = 5
\>>> i += 1
\>>> i
6
\>>> id(i)
140243713967984
\>>> i += 1
\>>> i
7
\>>> id(i)
140243713967960
可以看到执行 i += 1 时，内存地址都会变化，因为int 类型是不可变的。
再改改代码，但多个int类型的变量值相同时，看看它们内存地址是否相同。
\>>> i = 5
\>>> j = 5
\>>> id(i)
140656970352216
\>>> id(j)
140656970352216
\>>> k = 5
\>>> id(k)
140656970352216
\>>> x = 6
\>>> id(x)
140656970352192
\>>> y = 6
\>>> id(y)
140656970352192
\>>> z = 6
\>>> id(z)
140656970352192

对于不可变类型int，无论创建多少个不可变类型，只要值相同，都指向同个内存地址。同样情况的还有比较短
的字符串。

对于其他类型则不同，以浮点类型为例，从代码运行结果可以看出它是个不可变类型：对i的值进行修改后，指
向新的内存地址。
\>>> i = 1.5
\>>> id(i)
140675668569024
\>>> i = i + 1.7
\>>> i
3.2
\>>> id(i)
140675668568976
修改代码声明两个相同值的浮点型变量，查看它们的id，发现它们并不是指向同个内存地址，这点和int类型不
同（这方面涉及Python内存管理机制，Python对int类型和较短的字符串进行了缓存，无论声明多少个值相同
的变量，实际上都指向同个内存地址。）。
\>>> i = 2.5
\>>> id(i)
140564351733040
\>>> j = 2.5
\>>> id(j)
140564351733016 

可变类型的话，以list为例。list在append之后，还是指向同个内存地址，因为list是可变类型，可以在原处修
改。
\>>> a = [1, 2, 3]
\>>> id(a)
4385327224
\>>> a.append(4)
\>>> id(a)
4385327224
改改代码，当存在多个值相同的不可变类型变量时，看看它们是不是跟可变类型一样指向同个内存地址
\>>> a = [1, 2, 3]
\>>> id(a)
4435060856
\>>> b = [1, 2, 3]
\>>> id(b)
4435102392
从运行结果可以看出，虽然a、b的值相同，但是指向的内存地址不同。我们也可以通过b = a 的赋值语句，让
他们指向同个内存地址：
\>>> a = [1, 2, 3]
\>>> id(a)
4435060856
\>>> b = [1, 2, 3]
\>>> id(b)
4435102392
\>>> b = a
\>>> id(b)
4435060856
这个时候需要注意，因为a、b指向同个内存地址，而a、b的类型都是List，可变类型，对a、b任意一个List进
行修改，都会影响另外一个List的值。
\>>> b.append(4)
\>>> a
[1, 2, 3, 4]
\>>> b
[1, 2, 3, 4]
\>>> id(a)
4435060856
\>>> id(b)
4435060856
代码中，b变量append(4)，对a变量也是影响的。输出他们的内存地址，还是指向同个内存地址。         

​                                        __init__ 方法是什么？
__init__ 方法通常用在初始化一个类实例的时候，类似 java 等 OO 语言里的构造器。
例如：
class Person(object):
​    """Silly Person"""
​    def __init__(self, name, age):
​        self.name = name
​        self.age = age

​    def __str__(self):
​        return '<Person: %s(%s)>' % (self.name, self.age)

if __name__ == '__main__':
    piglei = Person('piglei', 24)
    print piglei
这样便是__init__最普通的用法了。但__init__其实不是实例化一个类的时候第一个被调用的方法。当使用 
Persion(name, age) 这样的表达式来实例化一个类时，最先被调用的方法其实是 __new__ 方法。 

__new__ 方法是什么？
__new__方法接受的参数虽然也是和__init__一样，但__init__是在类实例创建之后调用，而 __new__方法正
是创建这个类实例的方法。 
class Person(object):
    """Silly Person"""
    def __new__(cls, name, age):
        print '__new__ called.'
        return super(Person, cls).__new__(cls, name, age)

​    def __init__(self, name, age):
​        print '__init__ called.'
​        self.name = name
​        self.age = age

​    def __str__(self):
​        return '<Person: %s(%s)>' % (self.name, self.age)

if __name__ == '__main__':
    piglei = Person('piglei', 24)
    print piglei
执行结果： 
piglei@macbook-pro:blog$ python new_and_init.py 
__new__ called. 
__init__ called. 
<Person: piglei(24)>
通过运行这段代码，我们可以看到，__new__方法的调用是发生在__init__之前的。其实当 你实例化一个类的
时候，具体的执行逻辑是这样的：

p = Person(name, age)
首先执行使用name和age参数来执行Person类的__new__方法，这个__new__方法会 返回Person类的一个
实例（通常情况下是使用 super(Persion, cls).__new__(cls, ... ...) 这样的方式），
然后利用这个实例来调用类的__init__方法，上一步里面__new__产生的实例也就是 __init__里面的的 self
所以，__init__ 和 __new__ 最主要的区别在于： 

__init__ 通常用于初始化一个新实例，控制这个初始化的过程，比如添加一些属性， 做一些额外的操作，发生在类实例被创建完以后。它是实例级别的方法。
__new__ 通常用于控制生成一个新实例的过程。它是类级别的方法。
但是说了这么多，__new__最通常的用法是什么呢，我们什么时候需要__new__？ 

3、__new__ 的作用
依照Python官方文档的说法，__new__方法主要是当你继承一些不可变的class时(比如int, str, tuple)， 提供
给你一个自定义这些类的实例化过程的途径。还有就是实现自定义的metaclass。 

首先我们来看一下第一个功能，具体我们可以用int来作为一个例子： 

假如我们需要一个永远都是正数的整数类型，通过集成int，我们可能会写出这样的代码。 
class PositiveInteger(int):
    def __init__(self, value):
        super(PositiveInteger, self).__init__(self, abs(value))

i = PositiveInteger(-3)
print i
但运行后会发现，结果根本不是我们想的那样，我们任然得到了-3。这是因为对于int这种 不可变的对象，我们
只有重载它的__new__方法才能起到自定义的作用。 
这是修改后的代码： 
class PositiveInteger(int):
    def __new__(cls, value):
        return super(PositiveInteger, cls).__new__(cls, abs(value))

i = PositiveInteger(-3)
print i
通过重载__new__方法，我们实现了需要的功能。 
另外一个作用，关于自定义metaclass。其实我最早接触__new__的时候，就是因为需要自定义 metaclass，但鉴于篇幅原因，我们下次再来讲python中的metaclass和__new__的关系。 

4、用__new__来实现单例

事实上，当我们理解了__new__方法后，我们还可以利用它来做一些其他有趣的事情，比如实现 设计模式中的 单例模式(singleton) 。 
因为类每一次实例化后产生的过程都是通过__new__来控制的，所以通过重载__new__方法，我们 可以很简单的实现单例模式。 
class Singleton(object):
    def __new__(cls):
        \# 关键在于这，每一次实例化的时候，我们都只会返回这同一个instance对象
        if not hasattr(cls, 'instance'):
            cls.instance = super(Singleton, cls).__new__(cls)
        return cls.instance

obj1 = Singleton()
obj2 = Singleton()

obj1.attr1 = 'value1'
print obj1.attr1, obj2.attr1
print obj1 is obj2
输出结果： 
value1 value1 
True
可以看到obj1和obj2是同一个实例。

class Singleton(object):
    __instance = None

​    def __init__(self, *args, **kwargs):
​        pass

​    def __new__(cls, *args, **kwargs):
​        if not cls.__instance:
​            \# if not hasattr(cls, 'instance'):
​            cls.__instance = super(Singleton, cls).__new__(cls, *args, **kwargs)
​            cls.__instance.aa = args[0]
​            print type(cls), type(cls.__instance), type(cls.__instance.aa)
​        return cls.__instance

obj1 = Singleton(1, 2, 3, b=2)
obj2 = Singleton(1, 2, 3, b=2)

obj1.attr1 = 'value1'
obj2.attr2 = 'value2'
print obj1.attr1, obj1.attr2
print obj1 is obj2
print obj1.aa, obj2.attr1

结果：
<type 'type'> <class '__main__.Singleton'> <type 'int'>
value1 value2
True
1 value1

​                                        **类、实例和其他对象的内建函数**
\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**issubclass()**
issubclass() 布尔函数判断一个类是另一个类的子类或子孙类。
语法:
​    issubclass(sub, sup)

issubclass() 返回True的情况:
   \1.  给出的子类sub确实是父类sup的一个子类(反之,则为False)。 
   \2.  这个函数也允许"不严格"的子类,意味着,一个类可视为其自身的子类,所以,这个函数如果当 sub 就是 sup,或者从 sup 派生而来,则返回 True。(一个"严格的"子类是严格意义上的从一个类派生而来的子类。)
   \3.  从 Python 2.3 开始,issubclass()的第二个参数可以是可能的父类组成的 tuple(元组),这时, 只要第一个参数是给定元组中任何一个候选类的子类时,就会返回 True。

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**isinstance()**
isinstance() 布尔函数在判定一个对象是否是另一个给定类的实例时,非常有用。

语法:
    isinstance(obj1, obj2)
    isinstance()在 obj1 是类 obj2 的一个实例,或者是 obj2 的子类的一个实例时,返回 True (反之,则为 False)

例子:
    \>>> class C1(object): pass
    \>>> class C2(object): pass
    \>>> c1 = C1()
    \>>> c2 = C2()
    \>>> isinstance(c1, C1) True
    \>>> isinstance(c2, C1) False
    \>>> isinstance(c1, C2) False
    \>>> isinstance(c2, C2) True
    \>>> isinstance(C2, c2) 
    Traceback (innermost last):
    File "<stdin>", line 1, in ?
    isinstance(C2, c2)
    TypeError: second argument must be a class

注意:第二个参数应当是类;不然,你会得到一个 TypeError。但如果第二个参数是一个类型对象,则不会出现异常。这是允许的,因为你也可以使用 isinstance()来检查一个对象 obj1 是否是 obj2 的类型,比如:
    \>>> isinstance(4, int)
    True
    \>>> isinstance(4, str)
    False
    \>>> isinstance('4', str)
    True

isinstance()也可以使用一个元组(tuple)作为第二个参数。如果第一个参数是第二个参数中给定元组的任何一个候选类型或类的实例时,就会返回 True。

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**hasattr(), getattr(),setattr(), delattr()**
*attr()系列函数可以在各种对象下工作,不限于类(class)和实例(instances),因为在类和实例中使用极其频繁,就在这里列出来了。

hasattr()函数是 Boolean 型的,它的目的就是为了决定一个对象是否有一个特定的属性,一般用于访问某属性前先作一下检查。

getattr()取得对象的属性,会在你试图读取一个不存在的属性时,引发 AttributeError 异常,除非给出那个可选的默认参数。

setattr()将要么加入一个新的属性,要么取代一个已存在的属性。

delattr()函数会从一个对象中删除属性。

下面一些例子使用到了*attr()系列函数:
\>>> class myClass(object): 
...         def __init__(self): 
...             self.foo = 100 
\>>> myInst = myClass()
\>>> hasattr(myInst, 'foo')
True
\>>> getattr(myInst, 'foo')
100
\>>> hasattr(myInst, 'bar') 
False
\>>> getattr(myInst, 'bar') 
Traceback (most recent call last):
File "<stdin>", line 1, in ?
getattr(myInst, 'bar')
AttributeError: myClass instance has no attribute 'bar'
\>>> getattr(c, 'bar', 'oops!')
'oops!'
\>>> setattr(myInst, 'bar', 'my attr')
\>>> dir(myInst)
['__doc__', '__module__', 'bar', 'foo']
\>>> getattr(myInst, 'bar')       # same as myInst.bar 
'my attr'
\>>> delattr(myInst, 'foo')
\>>> dir(myInst)
['__doc__', '__module__', 'bar']
\>>> hasattr(myInst, 'foo')
False

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**dir()**
用 dir()列出一个模块所有属性的信息, dir()还可以用在对象上。
dir()提供的信息比以前更加详尽。根据文档,"除了实例变量名和常用方法外,它还显示那些通过特殊标记来调用的方法,像__iadd__(+=),__len__(len()), __ne__(!=)。
dir()作用在实例上(经典类或新式类)时,显示实例变量,还有在实例所在的类及所有它的基类中定义的方法和类属性.
dir()作用在类上(经典类或新式类)时,则显示类以及它的所有基类的__dict__中的内容。 但它不会显示定义在元类(metaclass)中的类属性.
dir()作用在模块上时,则显示模块的__dict__的内容.
dir()不带参数时,则显示调用者的局部变量.

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**super()**
super()函数的目的就是帮助程序员找出相应的父类, 然后方便调用相关的属性。一般情况下,程序员可能仅仅采用非绑定方式调用祖先类方法。使用 super()可以简化搜索一个合适祖先的任务,并且在调用它时,替你传入实例或类型对象。
每个定义的类,都有一个名为__mro__的属性,它是一个元组,按照他们被搜索时的顺序,列出了备搜索的类。
语法如下:
    super(type[, obj])
给出 type,super()"返回此 type 的父类"。如果你希望父类被绑定,你可以传入 obj 参数(obj 必须是 
type 类型的).否则父类不会被绑定。obj 参数也可以是一个类型,但它应当是 type 的一个子类。通常,当给出 obj 时:
   \1.  如果obj是一个实例,isinstance(obj,type)就必须返回True
   \2.  如果obj是一个类或类型,issubclass(obj,type)就必须返回True

事实上,super()是一个工厂函数,它创造了一个 super object,为一个给定的类使用__mro__ 去查找相应的父类。很明显,它从当前所找到的类开始搜索 MRO。
super()的主要用途,是来查找父类的属性,比如, super(MyClass,self).__init__()。如果你没有执行这样的查找,你可能不需要使用 super()。

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
**vars()**
vars()内建函数与 dir()相似,只是给定的对象参数都必须有一个__dict__属性。vars()返回一 个字典,它包含了对象存储于其__dict__中的属性(键)及值。如果提供的对象没有这样一个属性, 则会引发一个 TypeError 异常。如果没有提供对象作为 vars()的一个参数,它将显示一个包含本地名字空间的属性(键)及其值的字典,也就是locals()。

例子,使用类实例调用 vars():
    \>>> class C(object):
    ...         pass
    \>>> c = C()
    \>>> c.foo = 100
    \>>> c.bar = 'Python'
    \>>> c.__dict__
    {'foo': 100, 'bar': 'Python'}
    \>>> vars(c)
     {'foo': 100, 'bar': 'Python'}

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

​                                        **私有化**
默认情况下,属性在 Python 中都是"public",类所在模块和导入了类所在模块的其他模块的代码都可以访问到。很多 OO 语言给数据加上一些可见性,只提供访问函数来访问其值。这就是实现隐藏,是对象封装中的一个关键部分。
大多数 OO 语言提供"访问控制符"来限定成员函数的访问。

双下划线开头双下划线结尾的是一些 Python 的特殊对象，如类成员的 __init__、__del__、__add__、__getitem__等。 Python 官方推荐永远不要将这样的命名方式应用于自己的变量或函数
\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**双下划线(__)**
Python 为类元素(属性和方法)的私有性提供初步的形式。
由双下划线开始的属性在运行时被"混淆",所以直接访问是不允许的。实际上,会在名字前面加上下划线和类名。 
self.__num 属性为例:
    被"混淆"后,用于访问这个数据值的标识就变成了 self._NumStr__num。把类名加上后形成的新的"混淆"结果将可以防止在祖先类或子孙类中的同名冲突。
尽管这样做提供了某种层次上的私有化,但算法处于公共域中并且很容易被"击败"。这更多的是一种对导入源代码无法获得的模块或对同一模块中的其他代码的保护机制.

这种名字混淆的另一个目的,是为了保护__XXX 变量不与父类名字空间相冲突。如果在类中有一个__XXX 属性,它将不会被其子类中的__XXX 属性覆盖。(回忆一下,如果父类仅有一个 XXX 属性, 子类也定义了这个,这时,子类的 XXX 就是覆盖了父类的 XXX,这就是为什么你必须使用 PARENT.XXX 来调用父类的同名方法。) 
使用__XXX,子类的代码就可以安全地使用__XXX,而不必担心它会影响到父类中的__XXX。

例1：
In [14]: class A(object):
   ....:     __name = "wing"
   ....:     

In [15]: a=A()

In [16]: a.__name
AttributeError                            Traceback (most recent call last)
<ipython-input-16-5d5520ef9fe0> in <module>()
----> 1 a.__name
AttributeError: 'A' object has no attribute '__name'

In [18]: a._A__name   
Out[18]: 'wing'

例2：
In [30]: class C(object):
   ....:     name = "wing"  

In [31]: class D(C):
   ....:     __name = "tom"  

In [32]: d = D()

In [33]: d.name
Out[33]: 'wing'

In [34]: d._D__name
Out[34]: 'tom'

\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

**单下划线(_)**
简单的模块级私有化只需要在属性名前使用一个单下划线字符，防止模块的属性用"from mymodule 
import *"来加载。这是严格基于作用域的,所以这同样适合于函数。

假设test.py内容如下：
import  sys
_modulenum = 3
__modulenum = 3
__modulenum__ = 3
pubnum = 3
def  _fun():
       print 'this is',sys._getframe().f_code.co_name
def  __fun():
       print'this is ', sys._getframe().f_code.co_name
def  __fun__():
       print'this is ', sys._getframe().f_code.co_name
def  fun():
       print'this is ', sys._getframe().f_code.co_name

\>>> from  test import  *

\>>>_modulenum
...
NameError: name'_modulenum' is not defined

\>>>__modulenum
...
NameError: name'__modulenum' is not defined

\>>>__modulenum__
...
NameError: name'__modulenum__' is not defined

\>>>pubnum
3

\>>> _fun()
...
NameError: name'_fun' is not defined

\>>> __fun()
...
NameError: name'__fun' is not defined

\>>>__fun__()

...
NameError: name'__fun__' is not defined

\>>> fun()
this  is  fun

\>>> import test
\>>>test._modulenum
3
\>>>test.__modulenum
3
\>>>test.__modulenum__
3
\>>>test.pubnum
3
\>>>test._fun()
this  is  _fun
\>>>test.__fun()
this  is __fun
\>>>test.__fun__()
this  is __fun__
\>>>test.fun()
this  is  fun
\~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
注意，私有化都是针对外部而言，在类内部，依然可以使用正常的访问方式，在类的外部就必须进行"混淆"了。比如：
class test(object):
       __cversion =1.1
       def__init__(self):
              self.__iversion= 1.2
       deffun(self):
              printself.__iversion
              printtest.__cversion
              printself._test__iversion
              printself._test__cversion
\>>> from test import test
\>>> t1 = test()
\>>> t1.fun()
1.2
1.1
1.2
1.1
\>>> t1.__iversion
Traceback (most recent call last):
  File"<stdin>", line 1, in <module>
AttributeError: 'test' object has no attribute'__iversion'
\>>> t1._test__iversion
1.2
\>>> t1.__cversion
Traceback (most recent call last):
  File"<stdin>", line 1, in <module>
AttributeError: 'test' object has no attribute'__cversion'
\>>> t1._test__cversion
1.1

​                                        外翻

代码混淆 
代码混淆(Obfuscated code)亦称花指令，是将计算机程序的代码，转换成一种功能上等价，但是难于阅读和
理解的形式的行为。代码混淆可以用于程序源代码，也可以用于程序编译而成的中间代码。执行代码混淆的程
序被称作代码混淆器。目前已经存在许多种功能各异的代码混淆器。

将代码中的各种元素，如变量，函数，类的名字改写成无意义的名字。比如改写成单个字母，或是简短的无意
义字母组合，甚至改写成"__"这样的符号，使得阅读的人无法根据名字猜测其用途。重写代码中的部分逻辑，
将其变成功能上等价，但是更难理解的形式。比如将for循环改写成while循环，将循环改写成递归，精简中间
变量，等等。打乱代码的格式。比如删除空格，将多行代码挤到一行中，或者将一行代码断成多行等等。

代码混淆器也会带来一些问题。主要的问题包括：
被混淆的代码难于理解，因此调试以及除错也变得困难起来。开发人员通常需要保留原始的未混淆的代码用于
调试。对于支持反射的语言，代码混淆有可能与反射发生冲突。代码混淆并不能真正阻止反向工程，只能增大
其难度。因此，对于对安全性要求很高的场合，仅仅使用代码混淆并不能保证源代码的安全。

​                                        **Python新式类和经典类**
Python从2.2开始，引入了 new style class（新式类）
Python 2.x中默认都是经典类，只有显式继承了object才是新式类
Python 3.x中默认都是新式类，不必显式的继承object

新式类跟经典类的差别主要是以下几点:
• 新式类对象可以直接通过__class__属性获取自身类型:type 
• 继承搜索的顺序发生了改变
• 新式类增加了__slots__内置属性, 可以把实例属性的种类锁定到__slots__规定的范围之中
• 新式类增加了__getattribute__方法

\1. 新式类对象可以直接通过__class__属性获取自身类型:type 
\#cat test1.py
\#/usr/bin/env python2.7
    class E:    
    \#经典类  
        pass  
          
    class E1(object):    
    \#新式类  
        pass  
           
    e = E()  
    print "经典类"  
    print e  
    print type(e)  
    print e.__class__  
      
    print "新式类"  
    e1 = E1()  
    print e1  
    print e1.__class__  
    print type(e1)

输出结果：
    经典类  
    <__main__.E instance at 0x0000000002250B08>  
    <type 'instance'>  
    __main__.E  
      
    新式类  
    <__main__.E1 object at 0x0000000002248710>  
    <class '__main__.E1'>  
    <class '__main__.E1'> 

​    E1是定义的新式类。那么输出e1的时候，不论是type(e1)，还是e1.__class__都是输出的<class '__main__.E1'>。

\2. 继承搜索的顺序发生了改变
经典类多继承属性搜索顺序: 先深入继承树左侧，再返回，开始找右侧
新式类多继承属性搜索顺序: 先水平搜索，然后再向上移动
\#cat test2.py
    class A(object):    
        '新式类 ,作为所有类的基类'
        def foo(self):    
            print "class A"   
              
    class A1():    
        '经典类,作为所有类的基类 '
        def foo(self):    
            print "class A1"    
              
    class C(A):    
        pass  
          
    class C1(A1):    
        pass  
          
    class D(A):    
        def foo(self):    
            print "class D"    
          
    class D1(A1):    
        def foo(self):    
            print "class D1"    
      
    class E(C, D):    
        pass  
          
    class E1(C1, D1):    
        pass  
      
    e = E()  
    e.foo()  
             
    e1 = E1()  
    e1.foo() 

输出结果：
    class D  
    class A1  

因为A新式类,对于继承A类都是新式类，首先要查找类E中是否有foo()，如果没有则按顺序查找C->D->A。它是一种广度优先查找方式。

因为A1经典类,对于继承A1类都是经典类，首先要查找类E1中是否有foo()，如果没有则按顺序查找C1->A1->D1。它是一种深度优先查找方式。

\3. 新式类增加了__slots__内置属性, 可以把实例属性的种类锁定到__slots__规定的范围之中。

比如只允许对A实例添加name和age属性:
\#cat test3.py
    class A(object):    
        __slots__ = ('name', 'age')   
      
    class A1():    
        __slots__ = ('name', 'age')   
          
    a1 = A1()  
    a = A()  
      
    a1.name1 = "a1"  
    a.name1 = "a" 

A是新式类添加了__slots__ 属性,所以只允许添加 name age
A1经典类__slots__ 属性没用

执行结果：
     Traceback (most recent call last):  
       File "t.py", line 13, in <module>  
         a.name1 = "a"  
     AttributeError: 'A' object has no attribute 'name1'  
所以a.name是会出错的

通常每一个实例都会有一个__dict__属性，用来记录实例中所有的属性和方法，也是通过这个字典，可以让实例绑定任意的属性。
__slots__属性作用就是，当类C有比较少的变量，而且拥有__slots__属性时，类C的实例就没有__dict__属性，而是把变量的值存在一个固定的地方。如果试图访问一个__slots__中没有的属性，实例就会报错。
这样操作的好处：__slots__属性虽然令实例失去了绑定任意属性的便利，但是因为每一个实例没有__dict__属性，却能有效节省每一个实例的内存消耗，有利于生成小而精干的实例。

\4. 新式类增加了__getattribute__方法
\#cat test4.py
class A(object):    
    def __getattribute__(self, *args, **kwargs):    
        print "A.__getattribute__"  
                
class A1():    
    def __getattribute__(self, *args, **kwargs):    
        print "A1.__getattribute__"  
                
a1 = A1()  
a = A()  

a.test  
print "========="  
a1.test

A.__getattribute__  
=========  
Traceback (most recent call last):  
    File "t.py", line 18, in <module>  
    a1.test  
AttributeError: A1 instance has no attribute 'test'  

A是新式类，每次通过实例访问属性，都会经过__getattribute__函数,
A1不会调用__getattribute__所以出错了